/* ################ read_data.cpp #################################################
 *
 *
 */
#include "checkdae.h"

static int split_dae_file(char **fname, EqSystem_T *mySystem);
static int allocate_area(EqSystem_T *mySystem);
static int read_param_base(EqSystem_T *mySystem);
static int read_param(EqSystem_T *mySystem);
static int read_param_index(EqSystem_T *mySystem, FILE *fp_in);
static void read_param_data(EqSystem_T *mySystem, FILE *fp_in);
static int read_var(EqSystem_T *mySystem);
static int read_mid_expression(EqSystem_T *mySystem);
static int read_equation(EqSystem_T *mySystem, int f_isX);
static int fix_equation_line(char *fname_in, char *fname_out);
static int divide_line(char *line, char *strL, char *strM, char *strR);

/*---------------------- read_data -------------------------------------------
 *
 *	split_dae_file ǥǡ老Ȥ˥եʬ䤷
 *  ǡ老Ȥ˳ƥե뤫ǡɤ߹ࡣ
 *
 */
int
Read_data(EqSystem_T *mySystem, char **fname)
{
	int f_isX;
	
	/*
	 *  MathDAE.txt 饳Ȥ줾ǡ老Ȥ˥ƥݥե
	 *  񤭽Ф(head, usedp, const, var, mid, aeq, deq, param)
	 *  Ū롣
 	 */
	if (split_dae_file(fname, mySystem)) {
		log_sys("## Error in split_DAEfile. \n");
		goto	err;
	}

	/*
	 *  Ūåΰݤ롣
 	 */
	if (allocate_area(mySystem)) goto err;
	
	if (read_param_base(mySystem)) goto err;	/* ѥѥ᡼Υ٥ɤ߹ */

	if (read_param(mySystem)) goto err;	/* ġΥѥ᡼ɤ߹ */
	
	if (read_var(mySystem)) goto err;	/* ѿɤ߹ */

	if (read_mid_expression(mySystem)) goto err;	/* ɽɤ߹ */

	if (mySystem->n_varX > 0) {
		f_isX = true;
		if (read_equation(mySystem, f_isX)) goto err;	/* ɤ߹ */
	}

	f_isX = false;
	if (read_equation(mySystem, f_isX)) goto err;	/* ʬɤ߹ */

	return 0;

err:
	log_sys("## Error in read_data.\n");
	return -1;
}

/*---------------------- split_dae_file -------------------------------------------
 *
 *  MathDAE.txt 饳ȤäƤ줾ǡ老Ȥ˥ƥݥե
 *  񤭽Ф(head, usedp, const, var, mid, aeq, deq, param)
 *  ŪϿ롣
 *
 */
static int
split_dae_file(char **fname, EqSystem_T *mySystem)
{
	register char *cpt;
	register char *cpa;
	char fname_out[MAX_FNAME];
	char tline[BUFFSIZE];
	char aline[BUFFSIZE];
	int cnt, index=0, cnt_data=0, n_data[BLOCK_NUM];
	FILE *fp_in=NULL, *fp_out=NULL;

	memset(n_data, 0, sizeof(n_data));

	/* ϥե(MathDAE.txt)򥪡ץ󤹤 */
	if((fp_in = fopen(fname[FILE_D], "r")) == NULL){
		log_sys("#  Cannot open %s\n", fname[FILE_D]);
		goto err;
	}

	while(fgets(tline, sizeof tline, fp_in)) {	/* ǡɤ߹ߤΥ롼 */
		cpt = tline;	/* ե뤫ɤԤΥݥ */
		cpa = aline;	/* ̤Ȥ֤ԤΥݥ */

		cnt = 0;
		for (; *cpt; ) {	/* ϤΥ롼 */
			if (*cpt == '\n' | *cpt == '\r'|| *cpt == '\0' || *cpt == '#') {
				/* Ԥνλ */
				break;
			} else if (*cpt == ';') {
	   			*(cpa++) = *(cpt++);
				cnt++;
				cnt_data++;	/* ';'οǡб */
			} else {	/* ̾省饯ν񤭹 */
	   			*(cpa++) = *(cpt++);
				cnt++;
			}
		}

		if (cnt == 0) {	/* ͭʥǡʤ */
			continue;
		} else {		/* ǡä */
			*cpa   = '\0';
		}

		/* ϥեڤؤ */
		if (!strncmp(aline, "==", 2)) {
			n_data[index] = cnt_data;
			cnt_data = 0;
			if (!strncmp(aline, "== Mod", 6)) {
				strcpy(fname_out, FNAME_HEAD);
				index = 0;
			} else if (!strncmp(aline, "== Use", 6)) {
				strcpy(fname_out, FNAME_USEDP);
				index = 1;
			} else if (!strncmp(aline, "== Con", 6)) {
				strcpy(fname_out, FNAME_CONST);
				index = 2;
			} else if (!strncmp(aline, "== Var", 6)) {
				strcpy(fname_out, FNAME_VAR);
				index = 3;
			} else if (!strncmp(aline, "== Tem", 6)) {
				strcpy(fname_out, FNAME_MID);
				index = 4;
			} else if (!strncmp(aline, "== Alg", 6)) {
				strcpy(fname_out, FNAME_AEQ);
				index = 5;
			} else if (!strncmp(aline, "== Dif", 6)) {
				strcpy(fname_out, FNAME_DEQ);
				index = 6;
			} else if (!strncmp(aline, "== Par", 6)) {
				strcpy(fname_out, FNAME_PARAM);
				index = 7;
			} else {
				log_sys("#  Inappropriate line. \n\t%s\n", tline);
				goto err;
			}
			/* νϥե뤬Ĥ */
			if (fp_out != NULL) fclose(fp_out);
			/* ϥե򥪡ץ󤹤 */
			if((fp_out = fopen(fname_out, "w")) == NULL){
				log_sys("#  Cannot open %s\n", fname_out);
				goto err;
			}
		} else {
			if (fp_out == NULL) {
				log_sys("#  Inappropriate line. \n\t%s\n", tline);
				goto err;
			}
			fprintf(fp_out, "%s\n", aline);
		}

	}	/* End while */

	n_data[index] = cnt_data;

	if (fp_out != NULL) fclose(fp_out); fp_out = NULL;
	fclose(fp_in); fp_in = NULL;

	/*
	 * ŪͤϿ
	 */
	mySystem->n_param = n_data[1] + 1;	/* constantPlayer ʬ+1 */
	mySystem->n_const = n_data[2];
	mySystem->n_var   = n_data[3];
	mySystem->n_mid   = n_data[4];
	mySystem->n_varX  = n_data[5];
	mySystem->n_varY  = n_data[6];
	mySystem->n_equation  = n_data[5] + n_data[6];
	
	return 0;

err:
	log_sys("## Error in split_dae_file. \n");
	if (fp_out != NULL) fclose(fp_out);
	if (fp_in != NULL) fclose(fp_in); 
	return -1;

}

/*---------------------- allocate_area -------------------------------------------
 *
 *  Ūå¤ѤΥꥢݤ롣
 *
 */
static int
allocate_area(EqSystem_T *mySystem)
{
	if (mySystem->n_var < 1) {
		log_sys("#  There is no variable. \n");
		goto err;
	}
	if (mySystem->n_equation < 1) {
		log_sys("#  There is no equation. \n");
		goto err;
	}
	if (mySystem->n_equation > mySystem->n_var) {
		log_sys("#  Lack of variables. Number of equations : %d, Number of variables : %d\n",
						mySystem->n_equation, mySystem->n_var);
		goto err;
	}
	
	mySystem->var = (Variable_T *) calloc(mySystem->n_var, sizeof(Variable_T));
	if (mySystem->var == NULL) {
		log_sys("#  Can't allocate memory area for variables. \n");
		goto err;
	}
	mySystem->mid = (Expression_T *) calloc(mySystem->n_mid, sizeof(Expression_T));
	if (mySystem->mid == NULL) {
		log_sys("#  Can't allocate memory area for mid-expressions. \n");
		goto err;
	}
	mySystem->equation = (Expression_T *) calloc(mySystem->n_equation, sizeof(Expression_T));
	if (mySystem->mid == NULL) {
		log_sys("#  Can't allocate memory area for equations. \n");
		goto err;
	}
	mySystem->param_base = (ParamBase_T *) calloc(mySystem->n_param, sizeof(ParamBase_T));
	if (mySystem->mid == NULL) {
		log_sys("#  Can't allocate memory area for parameter labels. \n");
		goto err;
	}
	return 0;

err:
	log_sys("## Error in allocate_area. \n");
	return -1;

}

/*---------------------- read_param_base -------------------------------------------
 *
 *	split_dae_file ʬ䤷ѥѥ᡼̾ե(usedp)ǡɤ߹ࡣ
 *
 */
static int
read_param_base(EqSystem_T *mySystem)
{
	int cnt = 0;
  	unsigned int n_max = BUFFSIZE - 1;
	FILE *fp_in;
	char tline[BUFFSIZE];
	char str_label[BUFFSIZE], str_tmp1[BUFFSIZE], str_tmp2[BUFFSIZE];

	if((fp_in = fopen(FNAME_USEDP, "r")) == NULL){
		log_sys("#  Cannot open %s\n", FNAME_USEDP);
		return -1;
	}

	while(fgets(tline, sizeof tline, fp_in)) {
		if (strlen(tline) > n_max) {
			log_sys("#  Too long input line :'%s'\n", tline);
			goto err;
		}
		if (divide_line(tline, str_label, str_tmp1, str_tmp2)) goto err;
		if (strlen(str_tmp1) > 0 || strlen(str_tmp2) > 0) {
			log_sys("#  Inappropriate line.\n");
			goto err;
		}
		if (strlen(str_label) == 0) {
			log_sys("#  Lack of label.\n");
			goto err;
		}
		mySystem->param_base[cnt + 1].label = strdup(str_label);
		cnt++;
		if (cnt > mySystem->n_param) {
			log_sys("#  Too many parameter labels. Why??\n");
			goto err;
		}

	}	/* End while */

#if 0
	if (cnt == 0) {
		log_sys("#  There is no appropriate parameter label.\n");
		goto err;
	}
#endif

	mySystem->param_base[0].label = strdup(CONSTANT_PLAYER);

	fclose(fp_in); fp_in = NULL;

	return 0;

err:
	log_sys("#  in line : %s", tline);
	log_sys("## Error in read_param_base.\n");
	if (fp_in != NULL) fclose(fp_in);
	return -1;

}

/*---------------------- read_param -------------------------------------------
 *
 *	split_dae_file ʬ䤷ѥ᡼ե(const, param)
 *  read_param_index ˤꡢκ祤ǥåɤ߹ΰݤ
 *  read_param_dataˤɤ߹ǹ¤Τ˳Ǽ롣
 *
 */
static int
read_param(EqSystem_T *mySystem)
{
	int i, n_file;
	FILE *fp_in=NULL;
	char fname[2][MAX_FNAME];

	strcpy(fname[0], FNAME_PARAM);
	strcpy(fname[1], FNAME_CONST);
	
	if (mySystem->n_const == 0) {
		n_file = 1;
	} else {
		n_file = 2;
	}
	
	for (i=0; i<n_file; i++) {

		if((fp_in = fopen(fname[i], "r")) == NULL){
			log_sys("#  Cannot open %s\n", fname[i]);
			goto err;
		}

		if (read_param_index(mySystem, fp_in)) goto err;
		
		fclose(fp_in); fp_in = NULL;

	}	/* End of for */

	for (i=0; i<mySystem->n_param; i++) {
		mySystem->param_base[i].param = 
				(Parameter_T *) calloc(mySystem->param_base[i].max_index, sizeof(Parameter_T));
		if (mySystem->param_base[i].param == NULL) {
			log_sys("#  Can't allocate memory area for parameter '%s'. \n",
							mySystem->param_base[i].label);
			goto err;
		}
	}
		
	for (i=0; i<n_file; i++) {

		if((fp_in = fopen(fname[i], "r")) == NULL){
			log_sys("#  Cannot open %s\n", fname[i]);
			goto err;
		}

		read_param_data(mySystem, fp_in);
		
		fclose(fp_in); fp_in = NULL;

	}	/* End of for */

	return 0;

err:
	log_sys("## Error in read_param.\n");
	if (fp_in != NULL) fclose(fp_in);
	return -1;

}

/*---------------------- read_param_index -------------------------------------------
 *
 *	split_dae_file ʬ䤷ѥ᡼ե(constparam)
 *  ǥåκͤ롣
 *
 */
static int
read_param_index(EqSystem_T *mySystem, FILE *fp_in)
{
	int my_index, param_addr;
  	unsigned int n_max = BUFFSIZE - 1;
	char tline[BUFFSIZE];
	char str_var[BUFFSIZE], str_name[BUFFSIZE], str_index[BUFFSIZE],
   			str_val[BUFFSIZE], str_info[BUFFSIZE];

	while(fgets(tline, sizeof tline, fp_in)) {
		if (strlen(tline) > n_max) {
			log_sys("#  Too long input line :'%s'\n", tline);
			goto err;
		}
		if (divide_line(tline, str_var, str_val, str_info)) goto err;
		if (Divide_variable(str_var, str_name, str_index)) {
			log_sys("#  Inappropriate parameter: '%s'.\n", str_var);
			goto err;
		}
		if (strlen(str_index) == 0) {
			log_sys("#  Lack of index : '%s'.\n", str_var);
			goto err;
		}
		my_index = atoi(str_index);
		if (my_index == 0) {
			log_sys("#  Inappropriate parameter: '%s' in '%s'.\n", str_index, str_var);
			goto err;
		}
		param_addr = FindParamBaseAddr(str_name);
		if (param_addr < 0) {
			log_sys("#  '%s' is not registered as used parameter.\n", str_name);
			goto err;
		}
		if (my_index > mySystem->param_base[param_addr].max_index) {
			mySystem->param_base[param_addr].max_index = my_index;
		}

	}
		
	return 0;

err:
	log_sys("#  in line : %s", tline);
	log_sys("## Error in read_param_index.\n");
	return -1;

}

/*---------------------- read_param_data -------------------------------------------
 *
 *	split_dae_file ʬ䤷ѥ᡼ե(constparam)ǡɤ߹ࡣ
 *
 */
static void
read_param_data(EqSystem_T *mySystem, FILE *fp_in)
{
	int my_index, param_addr;
	char tline[BUFFSIZE];
	char str_var[BUFFSIZE], str_name[BUFFSIZE], str_index[BUFFSIZE],
   			str_val[BUFFSIZE], str_info[BUFFSIZE];

	while(fgets(tline, sizeof tline, fp_in)) {
		divide_line(tline, str_var, str_val, str_info);
		Divide_variable(str_var, str_name, str_index);
		my_index = atoi(str_index);
		param_addr = FindParamBaseAddr(str_name);
		mySystem->param_base[param_addr].param[my_index-1].tag = strdup(str_info);
		if (strlen(str_val) > 0) {
			mySystem->param_base[param_addr].param[my_index-1].val = atof(str_val);
		} else {
			mySystem->param_base[param_addr].param[my_index-1].val = PARAM_NULL;
		}
	}
}

/*---------------------- read_var -------------------------------------------
 *
 *	split_dae_file ʬ䤷ѿե(var)ǡɤ߹ࡣ
 *
 */
static int
read_var(EqSystem_T *mySystem)
{
	int i, j, cnt = 0, f_isX = false;
  	unsigned int n_max = BUFFSIZE - 1;
	FILE *fp_in;
	char tline[BUFFSIZE];
	char str_var[BUFFSIZE], str_name[BUFFSIZE], str_index[BUFFSIZE],
   			str_val[BUFFSIZE], str_info[BUFFSIZE];

	if((fp_in = fopen(FNAME_VAR, "r")) == NULL){
		log_sys("#  Cannot open %s\n", FNAME_VAR);
		return -1;
	}

	while(fgets(tline, sizeof tline, fp_in)) {
		if (strlen(tline) > n_max) {
			log_sys("#  Too long input line :'%s'\n", tline);
			goto err;
		}
		if (strncmp(tline, "= Alge", 6) == 0) {
			f_isX = true;
			continue;
		} else if (strncmp(tline, "= Diff", 6) == 0) {
			f_isX = false;
			continue;
		}
		if (divide_line(tline, str_var, str_val, str_info)) goto err;
		if (Divide_variable(str_var, str_name, str_index)) {
			log_sys("#  Inappropriate expression : '%s'.\n", str_var);
			goto err;
		}
		if (f_isX) {
			if (strcmp(str_name, "x") != 0) {
				log_sys("#  '%s' is not a variable of algebraic equations.\n", str_var);
				goto err;
			}
		} else {
			if (strcmp(str_name, "y") != 0) {
				log_sys("#  '%s' is not a variable of differential equations.\n", str_var);
				goto err;
			}
		}
		if (strlen(str_index) == 0) {
			log_sys("#  Lack of index : '%s'.\n", str_var);
			goto err;
		}
		mySystem->var[cnt].tag = strdup(str_index);
		mySystem->var[cnt].f_isX = f_isX;
		if (strlen(str_val) > 0) {
			mySystem->var[cnt].val = atof(str_val);
		} else {
			mySystem->var[cnt].val = -1;
		}
		cnt++;
		if (cnt > mySystem->n_var) {
			log_sys("#  Too many variables. Why??\n");
			goto err;
		}

	}	/* End while */

	if (cnt == 0) {
		log_sys("#  There is no appropriate variable.\n");
		goto err;
	}


	// ǥåνʣʤɤǧ
	for (i=0; i<cnt-1; i++) {
		for (j=i+1; j<cnt; j++) {
			if (strcmp(mySystem->var[i].tag, mySystem->var[j].tag) == 0) {
				log_sys("#  Duplicate name : '%s' : %d and %d\n",
							mySystem->var[i].tag, i+1, j+1);
				goto err;
			}
		}
	}

	fclose(fp_in); fp_in = NULL;

	return 0;

err:
	log_sys("#  in line : %s", tline);
	log_sys("## Error in read_var.\n");
	if (fp_in != NULL) fclose(fp_in);
	return -1;

}

/*---------------------- read_mid_expression -------------------------------------------
 *
 *	split_dae_file ʬ䤷ɽե(mid)ǡɤ߹ࡣ
 *
 */
static int
read_mid_expression(EqSystem_T *mySystem)
{
	int cnt = 0, f_isTotal = false;
  	unsigned int n_max = BUFFSIZE - 1;
	FILE *fp_in=NULL;
	char tline[BUFFSIZE];
	char str_name[BUFFSIZE], str_expr[BUFFSIZE], str_tmp[BUFFSIZE];

	if (mySystem->n_mid == 0) return 0;

	/*
	 *	ǲԤʤɤäƤ硢
	 *	ƥݥե˽񤭽Ф
	 */
	if (fix_equation_line(FNAME_MID, FNAME_TMP)) goto err;
	
	if((fp_in = fopen(FNAME_TMP, "r")) == NULL){
		log_sys("#  Cannot open %s\n", FNAME_TMP);
		return -1;
	}

	while(fgets(tline, sizeof tline, fp_in)) {
		if (strlen(tline) > n_max) {
			log_sys("#  Too long input line :'%s'\n", tline);
			goto err;
		}
		if (strncmp(tline, "= Tota", 6) == 0) {
			f_isTotal = true;
			continue;
		} else if (strncmp(tline, "= Flux", 6) == 0) {
			f_isTotal = false;
			continue;
		}
		if (divide_line(tline, str_name, str_expr, str_tmp)) goto err;
		if (strlen(str_tmp) > 0) {
			log_sys("#  Inappropriate line.\n");
			goto err;
		}
		if (strlen(str_name) == 0) {
			log_sys("#  Lack of name.\n");
			goto err;
		}
		if (strlen(str_expr) == 0) {
			log_sys("#  Lack of expression.\n");
			goto err;
		}
		mySystem->mid[cnt].name = strdup(str_name);
		if (str_expr[0] == '+') {
			mySystem->mid[cnt].expr = strdup(str_expr+1);
		} else {
			mySystem->mid[cnt].expr = strdup(str_expr);
		}
		mySystem->mid[cnt].f_isXorT = f_isTotal;
		cnt++;
		if (cnt > mySystem->n_mid) {
			log_sys("#  Too many mid-explessions. Why??\n");
			goto err;
		}

	}	/* End while */


	fclose(fp_in); fp_in = NULL;

	return 0;

err:
	log_sys("#  in line : %s", tline);
	log_sys("## Error in read_mid_expression.\n");
	if (fp_in != NULL) fclose(fp_in);
	return -1;

}

/*---------------------- read_equation -------------------------------------------
 *
 *	split_dae_file ʬ䤷ξե(aeq, deq)ǡɤ߹ࡣ
 *  f_isX=true 
 *  f_isX=falseʬ
 *
 */
static int
read_equation(EqSystem_T *mySystem, int f_isX)
{
	int cnt;
  	unsigned int n_max = BUFFSIZE - 1;
	FILE *fp_in=NULL;
	char tline[BUFFSIZE], fname_eq[MAX_FNAME];
	char strL[BUFFSIZE], strR[BUFFSIZE], str_tmp[BUFFSIZE];
	char str_name[BUFFSIZE], str_index[BUFFSIZE];

	*tline = '\0';

	if(f_isX){
		strcpy(fname_eq, FNAME_AEQ);
		cnt = 0;
	} else {
		strcpy(fname_eq, FNAME_DEQ);
		cnt = mySystem->n_varX;
	}
	
	/*
	 *	ǲԤʤɤäƤ硢
	 *	ƥݥե˽񤭽Ф
	 */
	if (fix_equation_line(fname_eq, FNAME_TMP)) goto err;
	
	if((fp_in = fopen(FNAME_TMP, "r")) == NULL){
		log_sys("#  Cannot open %s\n", FNAME_TMP);
		return -1;
	}

	while(fgets(tline, sizeof tline, fp_in)) {
		if (strlen(tline) > n_max) {
			log_sys("#  Too long input line :'%s'\n", tline);
			goto err;
		}
		if (divide_line(tline, strL, strR, str_tmp)) goto err;
		if (strlen(str_tmp) > 0) {
			log_sys("#  Inappropriate line.\n");
			goto err;
		}
		if (strlen(strL) == 0) {
			log_sys("#  Lack of left expression.\n");
			goto err;
		}
		if (strlen(strR) == 0) {
			log_sys("#  Lack of right expression.\n");
			goto err;
		}
		if (f_isX) {
			if (strcmp(strL, "0") == 0) {
				strcpy(str_tmp, strR);
			} else if (strcmp(strR, "0") == 0) {
				strcpy(str_tmp, strL);
			} else {
				if (strR[0] == '+') {
					sprintf(str_tmp, "%s-(%s)", strL, strR+1);
				} else {
					sprintf(str_tmp, "%s-(%s)", strL, strR);
				}
			}
			mySystem->equation[cnt].expr = strdup(str_tmp);
		} else {
			if (Divide_variable(strL, str_name, str_index)) {
				log_sys("#  Inappropriate expression : '%s'.\n", strL);
				goto err;
			}
			if (strcmp(str_name, "dydt") != 0) {
				log_sys("#  Inappropriate expression : '%s'.\n", str_name);
				goto err;
			}
			if (strlen(str_index) == 0) {
				log_sys("#  Lack of index.\n");
				goto err;
			}
			mySystem->equation[cnt].name = strdup(str_index);
			if (strR[0] == '+') {
				mySystem->equation[cnt].expr = strdup(strR+1);
			} else {
				mySystem->equation[cnt].expr = strdup(strR);
			}
			
		}
		mySystem->equation[cnt].f_isXorT = f_isX;
		cnt++;
		if (f_isX) {
			if (cnt > mySystem->n_varX) {
				log_sys("#  Too many algebraic equations. Why??\n");
				goto err;
			}
		} else {
			if (cnt > mySystem->n_varX + mySystem->n_varY) {
				log_sys("#  Too many differential equations. Why??\n");
				goto err;
			}
		}

	}	/* End while */


	fclose(fp_in); fp_in = NULL;

	return 0;

err:
	if (*tline) log_sys("#  in line : %s", tline);
	log_sys("## Error in read_equation.\n");
	if (fp_in != NULL) fclose(fp_in);
	return -1;
}

/*---------------------- fix_equation_line -------------------------------------------
 *
 *	˲ԤäƤԤԤ1Ԥˤ롣(File to File)
 *	aeq_*: ڤ deq_*: ƬäƤϤ
 */
static int
fix_equation_line(char *fname_in, char *fname_out)
{
	int  f_head, f_tail=true, f_tail_pre, cp_tail;
	unsigned int n_max = BUFFSIZE-1;
	FILE *fp_in, *fp_out;
	char tline[BUFFSIZE], aline[BUFFSIZE], *str_tmp;

	if((fp_in = fopen(fname_in, "r")) == NULL){
		log_sys("#  Cannot open %s\n", fname_in);
		return -1;
	}
	if((fp_out = fopen(fname_out, "w")) == NULL){
		log_sys("#  Cannot open %s\n", fname_out);
		return -1;
	}

	f_head = true;
	while(fgets(tline, sizeof tline, fp_in)) {
		if (strlen(tline) > n_max) {
			log_sys("#  Too long input line :'%s'\n", tline);
			goto err;
		}
		
		f_tail_pre = f_tail;
		
		/* ɽ '= Total''= Flux' ̵񤭽Ф add 2003.4.21 */
		if (f_head && strncmp(tline, "= ", 2) == 0) {
			fprintf(fp_out, "%s", tline);
			continue;
		}

		/*
		 *	ɤ߹Ԥιζ򡢥֡꥿
		 *	';' ̵ͭĴ١äϼνλǤʤϥ顼
		 */
		cp_tail = Trim_tail(tline);
		if (strstr(tline, ";")) {
			if (tline[cp_tail] == ';') {
				f_tail =true;
			} else {
				log_sys("#  This line has some character after ';'.\n");
				goto err;
			}
		} else {
			f_tail = false;
		}

		/*
		 *	Ƭ aeqdeq äϼ
		 */
		if (strncmp(tline, "aeq_", 4) == 0 || strncmp(tline, "deq_", 4) == 0) {
			if (f_head != true) {
				log_sys("#  'aeq' or 'deq' does not come in the middle of expression.\n");
				log_sys("#  Probably ';' has fallen out in the previous line.\n");
				goto err;
			}
			str_tmp = strstr(tline, ":");
			if (str_tmp == NULL) {
				log_sys("#  Lack of ':', before an equation starts.\n");
				goto err;
			}
			strcpy(tline, ++str_tmp);
		}

		/*
		 *	ե饰˱Ԥ
		 */
		if (f_head) {						/* Ԥϳ */
			if (f_tail) {						/* ԤϽü */
				if (f_tail_pre) {					/* Ԥϴ */
					fprintf(fp_out, "%s\n", tline);
				} else {							/* ԤϷ³ */
					goto err;
				}
			} else {							/* Ԥ̤ü */
				if (f_tail_pre) {					/* Ԥϴ */
					strcpy(aline, tline);
				} else {							/* ԤϷ³ */
					goto err;
				}
			}
		} else {							/* ԤϷ³ */
			if (f_tail) {						/* ԤϽü */
				if (f_tail_pre) {					/* Ԥϴ */
					goto err;
				} else {							/* ԤϷ³ */
					strcat(aline, tline);
					fprintf(fp_out, "%s\n", aline);
				}
			} else {							/* Ԥ̤ü */
				if (f_tail_pre) {					/* Ԥϴ */
					goto err;
				} else {							/* ԤϷ³ */
					strcat(aline, tline);
				}
			}
		}

		/*
		 *	ɹΤΥե饰Υåȡ
		 */
		if (f_tail) {
			f_head = true;
		} else {
			f_head = false;
		}
	}

	fclose(fp_in); fp_in = NULL;
	fclose(fp_out); fp_out = NULL;

	return 0;

err:
	log_sys("#  in line : %s\n", tline);
	log_sys("## Error in fix_equation_line.\n");
	if (fp_in != NULL) fclose(fp_in);
	if (fp_out != NULL) fclose(fp_out);
	return -1;
}

/*---------------------- divide_line -------------------------------------------
 *
 *	ǡԤζ'=', ';' ʬ䤷strL, strM, strRȤ֤
 *	';'1ĤʤƤϤʤʤ'='ϤʤƤ⤤2İʾϥ顼
 *
 */
static int
divide_line(char *line, char *strL, char *strM, char *strR)
{
	register char *cpi;
	register char *cpa;
	int cnt_equal=0, cnt_semi=0;

	strL[0] = strM[0] = strR[0] = '\0';

   	cpi = line;	/* ɤ߹ԤΥݥ */
	cpa = strL;	/* ʬʸؤΥݥ */
   	for (; *cpi; ) {	/* ϤΥ롼 */
		if (*cpi == '\n' || *cpi == '\0') {	/* Ԥνλ */
			*cpa = '\0';
			break;
		} else if (*cpi == ' ' || *cpi == '\t') {	/* ɤФ */
			cpi++;
		} else if (*cpi == '=') {	/* strM γ(strLν */
			cpi++;
			*cpa = '\0';
			cpa = strM;
			cnt_equal++;
			if (cnt_equal > 1) {
				log_sys("#  Duplication of '='.\n");
				goto err;
			}
		} else if (*cpi == ';') {	/* strR γ(strMν) */
			cpi++;
			*cpa = '\0';
			cpa = strR;
			cnt_semi++;
			if (cnt_semi > 1) {
				log_sys("#  Duplication of ';'.\n");
				goto err;
			}
		} else {	/* ¾Υ饯ν񤭹 */
			*(cpa++) = *(cpi++);
		}
   	}
	*cpa = '\0';

#ifdef DBG
	DBGW(30, ("line : %s", line));
	DBGW(30, ("L,M,R: '%s', '%s', '%s'\n", strL, strM, strR));
#endif

	if (cnt_semi == 0) {
		log_sys("#  Lack of ';'.\n");
		goto err;
	}
	
	return 0;

err:
	log_sys("## Error in divide_line.\n");
	return -1;
}
