/*
 * "R & B": Real-coded GAs & Bit-string GAs
 *
 * digenesis module
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "randb.h"
#include "logUtil.h"
#include "utl_alloc.h"
#include "utl_rand.h"

/*-----
 * NormalExecuteؿ
 *	̾ˡGA¹Ԥ
 */
int NormalExecute(GaParam_T *ga_param)
{
	int		i;
	int		num_parent;	/* Ƥο */
	int		num_island;	/* ο */
	int		island_pop;	/* θο */
	int		num_elite;	/* ꡼ȿ */
	int		generation;	/* generation */
	int		rv = EXIT_ERROR;
	GaIsland_T	*islands = NULL;		/*  */
#ifdef DBG
	DBGW(10, ("NormalExecute enter.\n"));
#endif /* DBG */
	
	num_parent = ga_param->crossoverParam.numParent;
	num_island = ga_param->gaTypeParam.numIsland;
	island_pop = ga_param->gaTypeParam.islandPopulation;
	num_elite  = ga_param->digenesisParam.numElite;
	
	/* ΰ */
	if((islands = GaIslandVectorAlloc(num_island)) == NULL) {
		goto out;
	}
	if(MakeImmigrationBuffer(ga_param, islands) == EXIT_ERROR){
		goto out;
	}
	
	/* ΤȯȽν */
#ifndef DBG
	SetSeedByTime();
#endif /* DBG */
	for(i = 0; i < num_island; i++){
		islands[i].population = ga_param->gaTypeParam.islandPopulation;
		GenerateInitialPopulation(&(islands[i]), 
					&(ga_param->gaEncodeParam), &(ga_param->variableRegion));
	}

	/* Υե뤬ꤵƤϡɤ߹ */ 
	if(strlen(ga_param->ioParam.individualFileName) > 0){
		if(SetIndividualFromFile(ga_param, islands) == EXIT_ERROR){
			goto out;
		}
	}
	
	for(i = 0; i < num_island; i++){
		if(IslandFitness(ga_param, &(islands[i]), 0, island_pop) == EXIT_ERROR){
			goto out;
		}
		SortByFitness(&(islands[i]));
	}
	RecordBestGene(0, islands, num_island, &(ga_param->bestGene));
	PrintBestGeneAtPresent(0, ga_param);
	if(ga_param->ioParam.outputFile != NULL){
		PrintFileAll(0, ga_param, islands);
	}
	if(ga_param->bestGene.gaGene.fitness > 
		ga_param->gaTypeParam.finValue){
		rv = EXIT_NORMAL;
		PrintResult(ga_param);
		goto out;
	}
	
	/* 1ϤޤΤ0夬Τ */
	for(generation = 1; 
		generation < ga_param->gaTypeParam.maxGeneration; generation++){
		/* ܽäܽ */
		if(ga_param->gaTypeParam.gaType == DGA 
			&& ga_param->gaTypeParam.immigrationSwitch != IMMIGRATION_OFF
			&& generation%(ga_param->gaTypeParam.immigrationInterval) == 0){
			if(Immigrate(ga_param, islands) == EXIT_ERROR){
				goto out;
			}
		}
		/* ä */
		if(ga_param->gaTypeParam.gaType == DIGA
			&& generation == ga_param->gaTypeParam.integrateGeneration){
			IntegrateIslands(ga_param, islands);
			AjustParamsAfterIntegrate(ga_param);
			num_island = 1;
		}

		/* ˤĤƤν */
		for(i = 0; i < num_island; i++){
			/* ȸ */
			if(ga_param->crossoverParam.crossoverType != NO_CROSSOVER){
				if(IslandCrossoverNormal(ga_param, &(islands[i]), 
							num_elite, island_pop) == EXIT_ERROR){
					goto out;
				}
			}
			/* Ѱ */
			if(ga_param->mutationParam.mutationType != NO_MUTATION){
				if(IslandMutation(ga_param, &(islands[i]),
							num_elite, island_pop) == EXIT_ERROR){
					goto out;
				}
			}
			/* Ŭ٤׻Ŭ٤ǥ */
			if(IslandFitness(ga_param, &(islands[i]), num_elite, island_pop) == EXIT_ERROR){
				goto out;
			}
			SortByFitness(&(islands[i]));
		}
		
		/* ¹ԤˤɸΤ¸ */
		RecordBestGene(generation, islands, num_island, &(ga_param->bestGene));

		/* ɸϤӥեؤν */
		PrintBestGeneAtPresent(generation, ga_param);
		if(ga_param->ioParam.outputFile != NULL){
			PrintFileAll(generation, ga_param, islands);
		}

		/* λȽ */
		if(ga_param->bestGene.gaGene.fitness > 
			ga_param->gaTypeParam.finValue){
			break;
		}
	}
	/* ̤ɽ */
	PrintResult(ga_param);
	rv = EXIT_NORMAL;
	
out:
	FreeImmigrationBuffer(ga_param, islands);
	if(islands != NULL) FreeGaIslandVector(islands);
	if(ga_param->ioParam.outputFile != NULL) 
		fclose(ga_param->ioParam.outputFile);
#ifdef DBG
	DBGW(10, ("NormalExecute exit(%d)\n", rv));
#endif /* DBG */
	
	return rv;
}

/*-----
 * MGGExecuteؿ
 *	MGGGA¹Ԥ
 */     
int MGGExecute(GaParam_T *ga_param)
{
	int		i;
	int		num_parent;		/* Ƥο */
	int		num_island;		/* ο */
	int		island_pop;		/* θο */
	int		generation;	 	/* generation */
	int		temp_mgg_size;	/* temp_mggǿ */
	int		rv = EXIT_ERROR;
	GaIsland_T	*islands = NULL;		/*  */

#ifdef DBG
	DBGW(10, ("MGGExecute enter.\n"));
#endif /* DBG */

	
	num_parent = ga_param->crossoverParam.numParent;
	num_island = ga_param->gaTypeParam.numIsland;
	island_pop = ga_param->gaTypeParam.islandPopulation;
	temp_mgg_size = 
		ga_param->digenesisParam.numMGGChildren + 2; /* +2ϸ򴹤Ƥο */


	if((islands = GaIslandVectorAlloc(num_island)) == NULL){
		goto out;
	}
	if(MakeImmigrationBuffer(ga_param, islands) == EXIT_ERROR){
		goto out;
	}
	
	/* ΤȯȽν */
#ifndef DBG
	SetSeedByTime();
#endif /* DBG */

	for(i = 0; i < num_island; i++){
		islands[i].population = ga_param->gaTypeParam.islandPopulation;
		GenerateInitialPopulation(&(islands[i]),
					&(ga_param->gaEncodeParam), &(ga_param->variableRegion));
	}
	
	/* Υե뤬ꤵƤϡɤ߹ */ 
	if(strlen(ga_param->ioParam.individualFileName) > 0){
		if(SetIndividualFromFile(ga_param, islands) == EXIT_ERROR){
			goto out;
		}
	}

	for(i = 0; i < num_island; i++){
		if(IslandFitness(ga_param, &(islands[i]), 0, island_pop) == EXIT_ERROR){
			goto out;
		}
		SortByFitness(&(islands[i]));
	}
	RecordBestGene(0, islands, num_island, &(ga_param->bestGene));
	PrintBestGeneAtPresent(0, ga_param);
	if(ga_param->ioParam.outputFile != NULL){
		PrintFileAll(0, ga_param, islands);
	}
	if(ga_param->bestGene.gaGene.fitness > 
		ga_param->gaTypeParam.finValue){
		rv = EXIT_NORMAL;
		PrintResult(ga_param);
		goto out;
	}
	
	/* 1ϤޤΤ0夬Τ */
	for(generation = 1; 
		generation < ga_param->gaTypeParam.maxGeneration; generation++){
		/* ܽäܽ */
		if(ga_param->gaTypeParam.gaType == DGA 
			&& ga_param->gaTypeParam.immigrationSwitch != IMMIGRATION_OFF
			&& generation%(ga_param->gaTypeParam.immigrationInterval) == 0){
			if(Immigrate(ga_param, islands) == EXIT_ERROR){
				goto out;
			}
		}
		/* ä */
		if(ga_param->gaTypeParam.gaType == DIGA
			&& generation == ga_param->gaTypeParam.integrateGeneration){
			IntegrateIslands(ga_param, islands);
			AjustParamsAfterIntegrate(ga_param);
			num_island = 1;
		}
		
		/* ˤĤƤν */
		for(i = 0; i < num_island; i++){
			/* MGGˤ */
			if(ga_param->crossoverParam.crossoverType != NO_CROSSOVER){
				if(IslandMggDigenesis(ga_param, &(islands[i])) == EXIT_ERROR){
					goto out;
				}
			}
			/* Ѱ */
			if(ga_param->mutationParam.mutationType != NO_MUTATION){
				if(IslandMutation(ga_param, &(islands[i]), 0, island_pop) == EXIT_ERROR){
					goto out;
				}
				/* Ŭ٤׻(Ѱ̵ξ硤Ŭ٤Ϸ׻) */
				if(IslandFitness(ga_param, &(islands[i]), 0, island_pop) == EXIT_ERROR){
					goto out;
				}
			}	
			SortByFitness(&(islands[i]));
		}

		/* ¹ԤˤɸΤ¸ */
		RecordBestGene(generation, islands, num_island, &(ga_param->bestGene));

		/* ɸϤӥեؤν */
		PrintBestGeneAtPresent(generation, ga_param);
		if(ga_param->ioParam.outputFile != NULL){
			PrintFileAll(generation, ga_param, islands);
		}

		/* λȽ */
		if(ga_param->bestGene.gaGene.fitness > 
			ga_param->gaTypeParam.finValue){
			break;
		}
	}

	/* ̤ɽ */
	PrintResult(ga_param);
	rv = EXIT_NORMAL;
	
out:
	FreeImmigrationBuffer(ga_param, islands);
	
	if(islands != NULL) FreeGaIslandVector(islands);
	if(ga_param->ioParam.outputFile != NULL) 
		fclose(ga_param->ioParam.outputFile);
#ifdef DBG
	DBGW(10, ("MGGExecute exit(%d)\n", rv));
#endif /* DBG */

	
	return rv;
}
