/*
 *"R & B": Real-coded GAs & Bit-string GAs
 *
 * fitness module
 */
#include <stdlib.h>

#include "randb.h"
#include "logUtil.h"
#include "utl_encode.h"
#include "utl_i2d.h"

#ifdef LIFE
#include "lsInterface.h"
#else /* LIFE */
#include "usrFunc.h"
#endif /* LIFE */

/*
 * Static variables
 */
static GaIsland_T *sortIsland;

/*
 * Prottype declarations of a static functios
 */

/*-----
 * comp_gene_fitnessؿ
 * 	2ĤθΤŬ٤Ӥ
 */	
static int comp_gene_fitness(const void *gene1, const void *gene2);

/***********************
 * Public functions    *
 ***********************/

/*-----
 * Fitnessؿ
 *	Ϳ줿ΤŬ٤֤
 *	͡Ŭ
 *	GaParam_T(I), Ŭ٤׻
 */
int Fitness(const GaParam_T *ga_param, const GaGene_T *ga_gene, double *fitness)
{
	int		rv;
	int		temp_array[NUM_VARIABLE_MAX];
	double	usrfunc_arg[NUM_VARIABLE_MAX];
#ifdef DBG
	DBGW(10, ("Fitness enter.\n"));
#endif /* DBG */
	
	switch(ga_param->gaEncodeParam.encodeType){
	case REAL:
#ifdef LIFE
		rv = LsFitness(ga_gene->gene_r, ga_param->gaEncodeParam.numVariable,
				ga_param->comLine.argc, ga_param->comLine.argv, fitness);
#else /* LIFE */
		rv = UsrFitness(ga_gene->gene_r, ga_param->gaEncodeParam.numVariable,
				ga_param->comLine.argc, ga_param->comLine.argv, fitness);
#endif /* LIFE */	
		break;
	case BINARY:
		BinaryDecode(ga_gene->gene_b, ga_param->gaEncodeParam.geneLength,
					ga_param->gaEncodeParam.quantumNumber, 
					ga_param->gaEncodeParam.numVariable, temp_array);
		Int2Double(&(ga_param->variableRegion), temp_array, usrfunc_arg, 
						ga_param->gaEncodeParam.numVariable);
#ifdef LIFE
		rv = LsFitness(usrfunc_arg, ga_param->gaEncodeParam.numVariable,
				ga_param->comLine.argc, ga_param->comLine.argv, fitness);
#else /* LIFE */
		rv = UsrFitness(usrfunc_arg, ga_param->gaEncodeParam.numVariable,
				ga_param->comLine.argc, ga_param->comLine.argv, fitness);
#endif /* LIFE */
		break;
	case GRAY:
		GrayDecode(ga_gene->gene_b, ga_param->gaEncodeParam.geneLength,
					ga_param->gaEncodeParam.quantumNumber, 
					ga_param->gaEncodeParam.numVariable, temp_array);
		Int2Double(&(ga_param->variableRegion), temp_array, usrfunc_arg, 
						ga_param->gaEncodeParam.numVariable);
#ifdef LIFE
		rv = LsFitness(usrfunc_arg, ga_param->gaEncodeParam.numVariable,
				ga_param->comLine.argc, ga_param->comLine.argv, fitness);
#else /* LIFE */
		rv = UsrFitness(usrfunc_arg, ga_param->gaEncodeParam.numVariable,
				ga_param->comLine.argc, ga_param->comLine.argv, fitness);
#endif /* LIFE */
		break;
	default: break;
	}

	if(rv != EXIT_NORMAL && rv != EXIT_ERROR){
		log_sys(" ## Error: Invalid return cord from usrFitness.\n");
		rv = EXIT_ERROR;
		goto out;
	}

out:
#ifdef DBG
	DBGW(10, ("Fitness exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * SortByFitnessؿ
 * 	Ŭ٤ˤꡤθΤ򥽡
 * 	͡EXIT_NORMAL
 * 	GaIsland_T*(I/0)
 */
int SortByFitness(GaIsland_T *island)
{
	int i;
#ifdef DBG
	DBGW(10, ("SortByFitness enter.\n"));
#endif /* DBG */

	for(i = 0; i < island->population; i++){
		island->sortedIndex[i] = i;
	}
	sortIsland = island;
		
	qsort(island->sortedIndex, island->population, sizeof(int), &comp_gene_fitness);

#ifdef DBG
	DBGW(10, ("SortByFitness exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}

/*-----
 * RecordBestGeneؿ
 * 	ޤǤκɸΤŬ٤ȸγκɸΤ
 * 	Ŭ٤ӤɤΤäθΤ¸
 * 	)ȸ˻Ȥ
 *	͡EXIT_NORMAL
 *	int(I), GaGene_T**(I), 
 *	      οint(I), ɸBestGene_T*(I/O)
 */
int RecordBestGene(int generation, GaIsland_T *islands, int num_island, BestGene_T *best_gene)
{
	int	i;
#ifdef DBG
	DBGW(10, ("RecordBestGene enter.\n"));
#endif /* DBG */

	for(i = 0; i < num_island; i++){
		if (best_gene->gaGene.fitness < 
				islands[i].residents[islands[i].sortedIndex[0]].fitness){
			best_gene->gaGene = 
				islands[i].residents[islands[i].sortedIndex[0]];
			best_gene->obtainedGeneration = generation;
		}
	}
#ifdef DBG
	DBGW(10, ("RecordBestGene exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}

/***********************
 * Static functions    *
 ***********************/

/*-----
 * comp_gene_fitnessؿ
 * 	2ĤθΤŬ٤Ӥ
 * 	͡ 1(gene1->fitness <  gene2->fitness)
 * 		 0(gene1->fitness == gene2->fitness)
 * 		-1(gene1->fitness >  gene2->fitness)
 * 	1GaGene_T*(I), 2GaGene_T*(I)
 */	
static int comp_gene_fitness(const void *gene1_no, const void *gene2_no)
{
	int			temp1_no, temp2_no;
	int rv;
	GaGene_T 	*temp1, *temp2;
#ifdef DBG
	DBGW(10, ("comp_gene_fitness enter.\n"));
#endif /* DBG */

	temp1_no = *(int *)gene1_no;
	temp2_no = *(int *)gene2_no;
	temp1 = &(sortIsland->residents[temp1_no]);
	temp2 = &(sortIsland->residents[temp2_no]);

	if(temp1->fitness < temp2->fitness){
		rv = 1;
	} else if(temp1->fitness == temp2->fitness){
		rv = 0;
	} else {
		rv = -1;
	}
#ifdef DBG
	DBGW(10, ("comp_gene_fitness exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}	
	
	
