/*
 * "R & B": Real-coded GAs & Bit-string GAs
 *
 * MPI module
 */
#include <stdlib.h>
#include <mpi.h>

#include "randb.h"
#include "gaMpi.h"
#include "logUtil.h"
#include "utl_alloc.h"

/*
 * Prottype declarations of a static function
 */

/*-----
 * mpiaint_vector_allocؿ
 * 	MPI_AIntΰݤ
 */
static MPI_Aint *mpiaint_vector_alloc(int size);

/*-----
 * mpidatatype_vector_allocؿ
 * 	MPI_Datatypeΰݤ
 */
static MPI_Datatype *mpidatatype_vector_alloc(int size);

/***********************
 * Public functions    *
 ***********************/

/*-----
 * GetMpiProcInfoؿ
 * 	ޥ/졼֤Ȥ˻
 *	ΥץΥ󥯾
 */
int GetMpiProcInfo(InfoMpiProc_T *mpi_info)
{
	int namelen;
	int	rv = EXIT_ERROR;
#ifdef DBG
	DBGW(10, ("GetMpiProcInfo enter.\n"));
#endif /* DBG */
	
	if(MPI_Comm_rank(MPI_COMM_WORLD, &(mpi_info->myProcRank)) != MPI_SUCCESS){
		log_sys(" #Error: MPI_Comm_rank(In GetMpiProcInfo).\n");
		goto out;
	}
	if(MPI_Comm_size(MPI_COMM_WORLD, &(mpi_info->numProcs)) != MPI_SUCCESS){
		log_sys(" #Error: MPI_Comm_size(In GetMpiProcInfo).\n");
		goto out;
	}
	if(MPI_Get_processor_name(mpi_info->hostname, &namelen) != MPI_SUCCESS){
		log_sys(" #Error: MPI_Get_processor_name(In GetMpiProcInfo).\n");
		goto out;
	}
	rv = EXIT_NORMAL;
	fprintf(stderr, "#####CPU %d/%d (%s)\n", 
		mpi_info->myProcRank, mpi_info->numProcs, mpi_info->hostname);

out:
#ifdef DBG
	DBGW(10, ("GetMpiProcInfo exit(%d)\n", rv));
#endif /* DBG */
	
	return rv;
}

/*-----
 * TypeParamStruct
 * 	ޥե뤫ɤ߹ѥ᡼
 * 	ۤ뤿ι¤
 */
#define NUM_BLOCKS	6
int TypeParamStruct(GaParam_T *ga_param, MPI_Datatype *mpi_param)
{
	int				rv = EXIT_ERROR;
	MPI_Aint		start_address;	/* ѥ᡼¤ΤƬɥ쥹 */
	MPI_Aint		address;		/* ƸΤƬ */
	int				*block_length = NULL;	/* ƥ֥åǿ */
	MPI_Aint		*displacement = NULL;	/* ƬγƸΤѰ */
	MPI_Datatype	*typelist = NULL;		/* ǤΥǡ */

#ifdef DBG
	DBGW(10, ("TypeParamStruct enter.\n"));
#endif /* DBG */


	if((block_length = IntVectorAlloc(NUM_BLOCKS)) == NULL){
		log_sys(" ## Error: IntVectorAlloc(In TypeParamStruct).\n");
		goto out;
	}
	if((displacement = mpiaint_vector_alloc(NUM_BLOCKS)) == NULL){
		log_sys(" ## Error: mpiaint_vector_alloc(In TypeParamStruct).\n");
		goto out;
	}
	if((typelist = mpidatatype_vector_alloc(NUM_BLOCKS)) == NULL){
		log_sys(" ## Error: mpidatatype_vector_alloc(In TypeParamStruct).\n");
		goto out;
	}

	if(MPI_Address(ga_param, &start_address) != MPI_SUCCESS){
		log_sys(" ## ERROR: MPI_Address(In TypeParamStruct1).\n");
		goto out;
	}
	displacement[0] = 0;
	block_length[0] = 13;
	typelist[0] = MPI_INT;

	if(MPI_Address(&(ga_param->gaTypeParam.finValue), &address)
		!= MPI_SUCCESS){
		log_sys(" ## ERROR: MPI_Address(In TypeParamStruct2).\n");
		goto out;
	}
	displacement[1] = address - start_address;
	block_length[1] = 2;
	typelist[1] = MPI_DOUBLE;

	if(MPI_Address(&(ga_param->digenesisParam), &address)
		!= MPI_SUCCESS){
		log_sys(" ## ERROR: MPI_Address(In TypeParamStruct3).\n");
		goto out;
	}
	displacement[2] = address - start_address;
	block_length[2] = 8;
	typelist[2] = MPI_INT;

	if(MPI_Address(&(ga_param->crossoverParam.alpha), &address)
		!= MPI_SUCCESS){
		log_sys(" ## ERROR: MPI_Address(In TypeParamStruct4).\n");
		goto out;
	}
	displacement[3] = address - start_address;
	block_length[3] = 2;
	typelist[3] = MPI_DOUBLE;
		
	if(MPI_Address(&(ga_param->mutationParam), &address)
		!= MPI_SUCCESS){
		log_sys(" ## ERROR: MPI_Address(In TypeParamStruct5).\n");
		goto out;
	}
	displacement[4] = address - start_address;
	block_length[4] = 1;
	typelist[4] = MPI_INT;

	if(MPI_Address(&(ga_param->mutationParam.mutationRate), &address)
		!= MPI_SUCCESS){
		log_sys(" ## ERROR: MPI_Address(In TypeParamStruct6).\n");
		goto out;
	}
	displacement[5] = address - start_address;
	block_length[5] = 2;
	typelist[5] = MPI_DOUBLE;

	/* ǡι */
	if(MPI_Type_struct(NUM_BLOCKS, block_length, 
			displacement, typelist, mpi_param) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Type_Struct(In TypeParamStruct).\n"); 
		goto out;
	}
	
	/* ǡ򥷥ƥ */
	if(MPI_Type_commit(mpi_param) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Type_commit(In TypeParamStruct).\n"); 
		goto out;
	}


	rv = EXIT_NORMAL;
out:
	if(block_length != NULL) FreeIntVector(block_length);
	if(displacement != NULL) free(displacement);
	if(typelist != NULL) free(typelist);

#ifdef DBG
	DBGW(10, ("TypeParamStruct exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * DistributeParamsؿ
 *	ޥΤ߻
 *	ե뤫ɤѥ᡼ƤΥ졼֤
 */
int DistributeParams(GaParam_T *ga_param, MPI_Datatype datatype)
{
	int		i;
	int		rv = EXIT_ERROR;
#ifdef DBG
	DBGW(10, ("DistributeParams enter.\n"));
#endif /* DBG */

	for(i = 1; i < ga_param->mpiInfo.numProcs; i++){
		if(MPI_Send(ga_param, 1, datatype, i, 
							NORMAL_TAG, MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In DistributeParams).\n");
			goto out;
		}
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("DistributeParams exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * AcceptParamsؿ
 *	졼֤Τ߻
 *	ޥۤѥ᡼
 *	ѥ᡼¤GaParam_T򹹿
 */
int AcceptParams(GaParam_T *ga_param, MPI_Datatype datatype)
{
	int			rv = EXIT_ERROR;
	MPI_Status	stat;
#ifdef DBG
	DBGW(10, ("AcceptParams enter.\n"));
#endif /* DBG */

	if(MPI_Recv(ga_param, 1, datatype, MPI_ANY_SOURCE,
					MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Recv(In AcceptParams).\n");
		goto out;
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("AcceptParams exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
		
/*-----
 * DistributeVariableRegionؿ
 * 	ޥΤ߻
 * 	ޥե뤫ɤ߹
 * 	ѿΰƤΥ졼֤
 */
int DistributeVariableRegion(GaParam_T *ga_param)
{
	int		i;
	int		rv = EXIT_ERROR;	
#ifdef DBG
	DBGW(10, ("DistributeVariableRegion enter.\n"));
#endif /* DBG */

	for(i = 1; i < ga_param->mpiInfo.numProcs; i++){
		if(MPI_Send(ga_param->variableRegion.minArray,
					ga_param->gaEncodeParam.numVariable, MPI_DOUBLE,
					i, NORMAL_TAG, MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In DistributeVariableRegion1).\n");
			goto out;
		}
		if(MPI_Send(ga_param->variableRegion.maxArray,
					ga_param->gaEncodeParam.numVariable, MPI_DOUBLE,
					i, NORMAL_TAG, MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In DistributeVariableRegion2).\n");
			goto out;
		}
		if(ga_param->gaEncodeParam.encodeType != REAL){
			if(MPI_Send(ga_param->variableRegion.delta,
					ga_param->gaEncodeParam.numVariable, MPI_DOUBLE,
					i, NORMAL_TAG, MPI_COMM_WORLD) != MPI_SUCCESS){
				log_sys(" ## Error: MPI_Send(In DistributeVariableRegion3).\n");
				goto out;
			}
		}
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("DistributeVariableRegion exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
		
/*-----
 * AcceptVariableRegionؿ
 * 	졼֤Τ߻
 * 	ޥۤѿΰǼ
 */
int AcceptVariableRegion(GaParam_T *ga_param)
{
	int			rv = EXIT_ERROR;
	MPI_Status	stat;
#ifdef DBG
	DBGW(10, ("AcceptVariableRegion enter.\n"));
#endif /* DBG */

	if(MPI_Recv(ga_param->variableRegion.minArray,
				ga_param->gaEncodeParam.numVariable, MPI_DOUBLE, MPI_ANY_SOURCE,
				MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Recv(In AcceptVariableRegion1).\n");
		goto out;
	}
	if(MPI_Recv(ga_param->variableRegion.maxArray,
				ga_param->gaEncodeParam.numVariable, MPI_DOUBLE, MPI_ANY_SOURCE,				MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Recv(In AcceptVariableRegion2).\n");
		goto out;
	}
	if(ga_param->gaEncodeParam.encodeType != REAL){
		if(MPI_Recv(ga_param->variableRegion.delta,
				ga_param->gaEncodeParam.numVariable, MPI_DOUBLE, MPI_ANY_SOURCE, 				MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Recv(In AcceptVariableRegion3).\n");
			goto out;
		}
	}

	ga_param->variableRegion.arraySize 
			=ga_param->gaEncodeParam.numVariable;
	
	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("AcceptVariableRegion exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
	
/*-----
 * TypeIslandStruct
 * 	åȤ줿ѥ᡼
 * 	MPI_Type_structˤꡢѤ
 * 	¤Τ
 */
int TypeIslandStruct(const GaParam_T *ga_param, GaIsland_T *island, MPI_Datatype *mpi_island)
{
	int				i, j;
	int				island_pop;		/* θο */
	int				num_var;
	int				gene_length;
	int				num_block;
	MPI_Aint		start_address;	/* Ƭɥ쥹 */
	MPI_Aint		address;		/* ƸΤƬ */
	int				rv = EXIT_ERROR;
	int				*block_length = NULL;	/* ƥ֥åǿ */
	MPI_Aint		*displacement = NULL;	/* ƬγƸΤѰ */
	MPI_Datatype	*typelist = NULL;		/* ǤΥǡ */

#ifdef DBG
	DBGW(10, ("TypeIslandStruct enter.\n"));
#endif /* DBG */
	
	num_var = ga_param->gaEncodeParam.numVariable;
	island_pop = ga_param->gaTypeParam.islandPopulation;
	gene_length = ga_param->gaEncodeParam.geneLength;
	
	if(ga_param->gaEncodeParam.encodeType == REAL){
		num_block = island_pop + 1;
	} else {
		num_block = 2 * island_pop + 1;
	}
	
	/* ΰ */
	if((block_length = IntVectorAlloc(num_block)) == NULL){
		log_sys(" ## Error: IntVectorAlloc(In TypeIslandStruct).\n");
		goto out;
	}
	if((displacement = mpiaint_vector_alloc(num_block)) == NULL){
		log_sys(" ## Error: mpiaint_vector_alloc(In TypeIslandStruct).\n");
		goto out;
	}
	if((typelist = mpidatatype_vector_alloc(num_block)) == NULL){
		log_sys(" ## Error: mpidatatype_vector_alloc(In TypeIslandStruct).\n");
		goto out;
	}
	
	/* ǤѰ̤ */
	displacement[0] = 0;
	block_length[0] = island_pop + 1;
	typelist[0] = MPI_INT;	
	if(MPI_Address(island, &start_address) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Address(In TypeIslandStruct).\n"); 
		goto out;
	}

	if(ga_param->gaEncodeParam.encodeType == REAL){
		for(i = 0; i < island_pop; i++){
			if(MPI_Address(&(island->residents[i]), &address) != MPI_SUCCESS){
				log_sys(" ## Error: MPI_Address(In TypeIslandStruct).\n"); 
				goto out;
			}
			displacement[i + 1] = address - start_address;
			block_length[i + 1] = num_var + 1;
			typelist[i + 1] = MPI_DOUBLE;
		}
	} else {
		for(i = 0, j = 0; i < island_pop; i++){
			if(MPI_Address(&(island->residents[i]), &address) != MPI_SUCCESS){
				log_sys(" ## Error: MPI_Address(In TypeIslandStruct).\n");
				goto out;
			}
			displacement[j + 1] = address - start_address;
			block_length[j + 1] = 1;
			typelist[j + 1] = MPI_DOUBLE;
			j++;
			if(MPI_Address(&(island->residents[i].gene_b[0]), &address) != MPI_SUCCESS){
				log_sys(" ## Error: MPI_Address(In TypeIslandStruct).\n");
				goto out;
			}
			displacement[j + 1] = address - start_address;
			block_length[j + 1] = gene_length;
			typelist[j + 1] = MPI_CHAR;
			j++;
		}
	}
			
	/* ǡι */
	if(MPI_Type_struct(num_block, block_length, 
			displacement, typelist, mpi_island) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Type_Struct(In TypeIslandStruct).\n"); 
		goto out;
	}
	
	/* ǡ򥷥ƥ */
	if(MPI_Type_commit(mpi_island) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Type_Commit(In TypeIslandStruct).\n"); 
		goto out;
	}

	rv = EXIT_NORMAL;
out:
	if(block_length != NULL) FreeIntVector(block_length);
	if(displacement != NULL) free(displacement);
	if(typelist != NULL) free(typelist);

#ifdef DBG
	DBGW(10, ("TypeIslandStruct exit(%d)\n", rv));
#endif /* DBG */


	return rv;
}

/*-----
 * SendIslandؿ
 *	ޥ/졼֤Ȥ˻
 * 	ꤵ줿ꤵ줿ץ
 */
int SendIsland(GaIsland_T *island, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm)
{
	int		rv = EXIT_ERROR;
#ifdef DBG
	DBGW(10, ("SendIsland enter.\n"));
#endif /* DBG */

	if(MPI_Send(island, 1, datatype, dest, tag, comm) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Send(In SendIsland).\n");
		goto out;
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("SendIsland exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
		
/*-----
 * RecvIslandؿ
 *	ޥ/졼֤Ȥ˻
 * 	1ǻꤵ줿Хåե˥ԡ
 */
extern int RecvIsland(GaIsland_T *buf, MPI_Datatype datatype, int src, int tag, MPI_Comm comm, MPI_Status *stat)
{
	int		rv = EXIT_ERROR;
#ifdef DBG
	DBGW(10, ("RecvIsland enter.\n"));
#endif /* DBG */

	if(MPI_Recv(buf, 1, datatype, src, tag, comm, stat) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Recv(In RecvIsland).\n");
		goto out;
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("RecvIsland exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/***********************
 * Static functions    *
 ***********************/

/*-----
 * mpiaint_vector_allocؿ
 * 	MPI_Aintΰݤ
 */
static MPI_Aint *mpiaint_vector_alloc(int size)
{
	MPI_Aint	*temp;
#ifdef DBG
	DBGW(10, ("mpiaint_vector_alloc enter.\n"));
#endif /* DBG */

	temp = (MPI_Aint *)malloc(size * sizeof(MPI_Aint));
	if(temp == NULL){
		log_sys(" ## Error: Cannot allocate memory(In mpiaint_vector_alloc).\n");
	}
#ifdef DBG
	DBGW(10, ("mpiaint_vector_alloc exit\n"));
#endif /* DBG */

	return temp;
}

/*-----
 * mpidatatype_vector_allocؿ
 * 	MPI_Datatypeΰݤ
 */
static MPI_Datatype *mpidatatype_vector_alloc(int size)
{
	MPI_Datatype	*temp;
#ifdef DBG
	DBGW(10, ("mpidatatype_vector_alloc enter.\n"));
#endif /* DBG */

	temp = (MPI_Datatype *)malloc(size * sizeof(MPI_Datatype));
	if(temp == NULL){
		log_sys(" ## Error: Cannot allocate memory(In mpidatatype_vector_alloc).\n");
	}
#ifdef DBG
	DBGW(10, ("mpidatatype_vector_alloc exit\n"));
#endif /* DBG */


	return temp;
}


