/*
 *"R & B": Real-coded GAs & Bit-string GAs
 *
 * IO module
 */

#include <stdio.h>
#include <string.h>

#include "randb.h"
#include "logUtil.h"
#include "utl_alloc.h"
#include "utl_encode.h"
#include "utl_i2d.h"
#include "utl_text.h"

/* 
 * Prottype declaritions of a static function
 */

/*-----
 * print_read_summaryؿ
 *  ե뤫ɤ߹Τοʤɤɽ
 */
static int print_read_summary(FILE *fp, char *file_name, int *read_island_pop, int num_island, int num_read_island, int island_pop);

/*-----
 * print_fitness
 */
static int print_fitness(const GaGene_T *ga_gene, FILE *fp);

/*-----
 * print_ga_geneؿ
 * 	ؿǻꤵ줿Τξɽ
 */
static int print_ga_gene(GaGene_T *ga_gene, GaParam_T *ga_param, FILE *fp);

/***********************
 * Public functions    *
 ***********************/

/*-----
 * SetIndividualFromFileؿ
 *  Τե뤫ɤ߹
 *  : ｪλ(EXIT_NORMAL) 顼(EXIT_ERROR)
 *  : GaParam_T*(I), GaIsland_T*(O)
 */ 
int SetIndividualFromFile(GaParam_T *ga_param, GaIsland_T *islands)
{
	int i;
	int num_var;
	int num_island;
	int island_pop;
	int rv = EXIT_ERROR;
	int individual_no = 0;
	int island_no = 0;
	char *read_top;
	char buf[BUFSIZ];
	FILE *fp;
	int *read_island_pop;
	int	transform_buf[NUM_VARIABLE_MAX];
	double readin_buf[NUM_VARIABLE_MAX];
	
#ifdef DBG
	DBGW(10, ("SetIndividualFromFile enter.\n"));
#endif /* DBG */

	
	num_var = ga_param->gaEncodeParam.numVariable;
	num_island = ga_param->gaTypeParam.numIsland;
	island_pop = ga_param->gaTypeParam.islandPopulation;

	if((fp = fopen(ga_param->ioParam.individualFileName, "r")) == NULL){
		log_sys(" ## Error: Cannot open a file of initial individuals:%s\n",
						ga_param->ioParam.individualFileName);
		goto out;
	}

	if((read_island_pop = IntVectorAlloc(num_island)) == NULL){
		goto out;
	}
	memset(read_island_pop, 0, num_island * sizeof(int));
	
	while(fgets(buf, sizeof(buf), fp) != NULL){
		/* ȹԤξ̵ */ 
		if (IsCommentLine(buf) == COMMENT_LINE){
			continue;
		}

		/*
		 * 롼()ζڤ꤬ä
		 * ꤵ줿οʲɤ߹οϿɤ߹ֹ򹹿
		 * ꤵ줿礭ֿΤ߷׿
		 */
		if (!strcmp(buf,"**\n")){
			if(island_no < num_island){
				read_island_pop[island_no] = individual_no;
			}
			island_no++;
			individual_no = 0;
			continue;
		}

		/* ξɤ߹ */		
		if(island_no < num_island && individual_no < island_pop){
			/* ǽ2ϻѤʤ */
			if(strtok(buf, ",\n") == NULL || strtok(NULL, ",\n") == NULL){
				log_sys(" ## Error: Invalid records in a file of initial individuals.\n");
				goto out;
			}

			for(i = 0; (read_top = strtok(NULL, ",\n")) != NULL ; i++){
				if(i < ga_param->gaEncodeParam.numVariable) {
					readin_buf[i] = atof(read_top);
					if(readin_buf[i] < ga_param->variableRegion.minArray[i] ||
						readin_buf[i] > ga_param->variableRegion.maxArray[i]){
						log_sys(" ## Error: Out of region in a file of initial individuals.\n");
						goto out;
					}
				}
			}
			
			if(i != num_var){
				log_sys(" ## Error: Invalid dimensions of a file of initial individuales.\n");
				goto out;
			}

			if(ga_param->gaEncodeParam.encodeType == REAL){ /* RGA */
				/* ɤ߹ͤ򤽤Τޤޥԡ */
				memcpy(islands[island_no].residents[individual_no].gene_r,
							readin_buf, sizeof(double) * num_var);
			} else if(ga_param->gaEncodeParam.encodeType == BINARY){ /* BGA(Binary) */
				/* ¿ͤͤѴ */
				Double2Int(&(ga_param->gaEncodeParam), &(ga_param->variableRegion), readin_buf, transform_buf, num_var);
				/* ͤХʥꥳǥ󥰤ƽ񤭹 */
				if(BinaryEncode(transform_buf, num_var, 
						islands[island_no].residents[individual_no].gene_b,
						ga_param->gaEncodeParam.geneLength,
						ga_param->gaEncodeParam.quantumNumber) == EXIT_ERROR){
					goto out;
				}
			} else if(ga_param->gaEncodeParam.encodeType == GRAY){ /* BGA(Gray) */
				/* ¿ͤͤѴ */
				Double2Int(&(ga_param->gaEncodeParam), &(ga_param->variableRegion), readin_buf, transform_buf, num_var);

				/* ͤ򥰥쥤ǥ󥰤ƽ񤭹 */
				if(GrayEncode(transform_buf, num_var, 
						islands[island_no].residents[individual_no].gene_b,
						ga_param->gaEncodeParam.geneLength,
						ga_param->gaEncodeParam.quantumNumber) == EXIT_ERROR){
					goto out;
				}
			} else {
				log_sys(" ## System Error: In SetIndividualFromFile.\n");
				goto out;
			}
		}
		individual_no++;
	}
	if(island_no <= num_island){
		read_island_pop[island_no] = individual_no;
	}
	island_no++;
	
	/* ɤ߹󥵥ޥ꡼ɽ */
	print_read_summary(stdout, ga_param->ioParam.individualFileName, read_island_pop,
					num_island, island_no, island_pop);
	
	rv = EXIT_NORMAL;
out:

#ifdef DBG
	DBGW(10, ("SetIndividualFromFile exit(%d).\n", rv));
#endif /* DBG */

	return rv;
}



/*-----
 * PrintExecuteParametersؿ
 *	GA¹ԤݤΥѥ᡼ꤵ줿ȥ꡼˽
 *	μ¹ԤѤѥ᡼Τɽ
 *	͡EXIT_NORMAL
 *	GaParam_T*(I), FILE*(O)
 */
int PrintExecuteParameters(const GaParam_T *ga_param, FILE *fp)
{
	int	encode_type;
	int	ga_type;	/* ¹Է */
	int	digenesis; 	/* ˡ */
	int	crossover;	/* ˡ */
	int	mutation;	/* Ѱˡ */
	int	selection;	/* ˡ */
	int 	immigration;	/* ܽå */
	int	num_var;	/* ѿο(õ֤μ) */
	int	i;	
	
	char	encode_type_name[MAXNAMLEN];
	char	ga_type_name[MAXNAMLEN];
	char	digenesis_name[MAXNAMLEN];
	char	crossover_name[MAXNAMLEN];
	char	mutation_name[MAXNAMLEN];
	char	selection_name[MAXNAMLEN];
	char	immigration_name[MAXNAMLEN];

#ifdef DBG
	DBGW(10, ("PrintExecuteParameters enter.\n"));
#endif /* DBG */
	
	encode_type = ga_param->gaEncodeParam.encodeType;
	ga_type = ga_param->gaTypeParam.gaType;
	digenesis = ga_param->digenesisParam.digenesisType;
	selection = ga_param->digenesisParam.selectionType;
	crossover = ga_param->crossoverParam.crossoverType;
	mutation = ga_param->mutationParam.mutationType;
	immigration = ga_param->gaTypeParam.immigrationSwitch;
	num_var = ga_param->gaEncodeParam.numVariable;

	switch(encode_type){
	case REAL:
		strcpy(encode_type_name, "Real-coded");
		break;
	case BINARY:
		strcpy(encode_type_name, "Bit-string(Binary coding)");
		break;
	case GRAY:
		strcpy(encode_type_name, "Bit-string(Gray coding)");
		break;
	default:break;
	}
	
	switch(ga_type){
	case DGA:
		strcpy(ga_type_name, "DGA");
		switch (immigration){
		case RANDOM_SWAP:
			strcpy(immigration_name, "Random Swap");
			break;
		case RANDOM_RING:
			strcpy(immigration_name, "Random Ring");
			break;
		case IMMIGRATION_OFF:
			strcpy(immigration_name, "OFF");
			break;
		default: break;
		}
		break;
	case DIGA:
		strcpy(ga_type_name, "DIGA");
		break;
	default: break;
	}

	switch(digenesis){
	case NORMAL_DIGENESIS:
		strcpy(digenesis_name, "NORMAL");
		switch(selection) {
		case ROULETTE:
			strcpy(selection_name, "Roulette");
			break;
		case TOURNAMENT:
			strcpy(selection_name, "Tournament");
			break;
		case RANDOM:
			strcpy(selection_name, "Random");
			break;
		default: break;
		}
		break;
	case MGG:
		strcpy(digenesis_name, "MGG");
		break;
	default: break;
	}

	switch(crossover){
	case BLX:
		strcpy(crossover_name, "BLX");
		break;
	case UNDX:
		strcpy(crossover_name, "UNDX");
		break;
	case UNDXM:
		strcpy(crossover_name, "UNDXm");
		break;
	case SPX:
		strcpy(crossover_name, "SPX");
		break;
	case NPOINTS:
		strcpy(crossover_name, "N points crossover");
		break;
	case NO_CROSSOVER:
		strcpy(crossover_name, "None");
		break;
	default: break;
	}

	switch(mutation){
	case REGION_UNI:
		strcpy(mutation_name, "RegionUni");
		break;
	case FIXED_UNI:
		strcpy(mutation_name, "FixedUni");
		break;
	case FIXED_NORMAL:
		strcpy(mutation_name, "FixedNormal");
		break;
	case VARIABLE_UNI:
		strcpy(mutation_name, "VariableUni");
		break;
	case VARIABLE_NORMAL:
		strcpy(mutation_name, "VariableNoraml");
		break;
	case BIT_REVERSE:
		strcpy(mutation_name, "BitReverse");
		break;
	case NO_MUTATION:
		strcpy(mutation_name, "None");
		break;
	default: break;
	}
		

	fprintf(fp, "===GA PARAMETERS========================\n");
	fprintf(fp, "Encode Type          : %s\n", encode_type_name);
	if(encode_type != REAL){
		fprintf(fp, "Quantum number       : %d\n", ga_param->gaEncodeParam.quantumNumber);
		fprintf(fp, "Gene length          : %d\n", ga_param->gaEncodeParam.geneLength);
	}
	fprintf(fp, "\nGA type              : %s\n", ga_type_name);
	fprintf(fp, "Max Generation       : %d\n", ga_param->gaTypeParam.maxGeneration);
	fprintf(fp, "Fin value            : %e\n", ga_param->gaTypeParam.finValue);
	fprintf(fp, "Total population     : %d\n", ga_param->gaTypeParam.totalPopulation);
	fprintf(fp, "Number of islands    : %d\n", ga_param->gaTypeParam.numIsland);
	fprintf(fp, "Island population    : %d\n", ga_param->gaTypeParam.islandPopulation);
	fprintf(fp, "\nDigenesis type       : %s\n", digenesis_name);
	switch(digenesis){
	case NORMAL_DIGENESIS:
		fprintf(fp, "Number of elites     : %d\n", 
			ga_param->digenesisParam.numElite);
		fprintf(fp, "Selection type       : %s\n", selection_name);
		if(selection == TOURNAMENT){
			fprintf(fp, "Tournament Size      : %d\n",
				ga_param->digenesisParam.tournamentSize);
		}
		break;
	case MGG:
		fprintf(fp, "Number of children   : %d\n",
			ga_param->digenesisParam.numMGGChildren);
		break;
	default: break;
	}

	fprintf(fp, "\nCrossover type       : %s\n", crossover_name);
	switch(crossover) {
	case BLX:
		fprintf(fp, "BLX(alpha)           : %f\n",
			ga_param->crossoverParam.alpha);
		break;
	case UNDX:
		fprintf(fp, "UNDX(alpha)          : %f\n",
			ga_param->crossoverParam.alpha);
		fprintf(fp, "UNDX(beta)           : %f\n",
			ga_param->crossoverParam.beta);
		break;
	case UNDXM:
		fprintf(fp, "UNDXm(alpha)         : %f\n",
			ga_param->crossoverParam.alpha);
		fprintf(fp, "UNDXm(beta)          : %f\n",
			ga_param->crossoverParam.beta);
		fprintf(fp, "UNDXm(m)             : %d\n",
			ga_param->crossoverParam.m);
		break;
	case SPX:
		fprintf(fp, "SPX(epsilon)         : %f\n",
			ga_param->crossoverParam.alpha);
		break;
	case NPOINTS:
		fprintf(fp, "N points(n)          : %d\n",
			ga_param->crossoverParam.m);
		break;
	default: break;
	}

	fprintf(fp, "\nMutation type        : %s\n", mutation_name);
	if(mutation != NO_MUTATION) {
		fprintf(fp, "Mutation rate        : %f\n", 
			ga_param->mutationParam.mutationRate);
		if(mutation == FIXED_UNI){
			fprintf(fp, "Fixed width          : %f\n",
				ga_param->mutationParam.fixedWidth);
		}
		if(mutation == FIXED_NORMAL){
			fprintf(fp, "SD for FixedNormal   : %f\n",
				ga_param->mutationParam.fixedWidth);
		}
	}

	switch(ga_type){
	case DGA:
		fprintf(fp, "\nImmigration Switch   : %s\n", immigration_name);	
		if(immigration != IMMIGRATION_OFF){
			fprintf(fp, "Immigration interval : %d\n",
				ga_param->gaTypeParam.immigrationInterval);
			fprintf(fp, "Immigration rate     : %f\n",
				ga_param->gaTypeParam.immigrationRate);
		}
		break;
	case DIGA:
		fprintf(fp, "\nIntegrate Generation : %d\n",
			ga_param->gaTypeParam.integrateGeneration);
		break;
	default: break;
	}
			
	fprintf(fp, "\n\n===Searching space information==========\n");	
	fprintf(fp, "Number of Variable   : %d\n", num_var);
	for(i = 0; i < num_var; i++){
		if(ga_param->gaEncodeParam.encodeType == REAL) {
			fprintf(fp, "Domain variable[%2d]   : [%f, %f]\n",
				i, ga_param->variableRegion.minArray[i],
				ga_param->variableRegion.maxArray[i]);
		} else {	
			fprintf(fp, "Domain variable[%d]  : [%f, %f](%f)\n",
				i, ga_param->variableRegion.minArray[i],
				ga_param->variableRegion.maxArray[i],
				ga_param->variableRegion.delta[i]);
		}
	}
	fprintf(fp, "========================================\n");	

#ifdef DBG
	DBGW(10, ("PrintExecuteParameters exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}
	
/*-----
 * PrintBestGeneAtPresentؿ
 *	λǤΡɸΤɸϤɽ
 *	͡EXIT_NORMAL
 *	generation(I), GaParam_T*(I)
 */
int PrintBestGeneAtPresent(int generation, GaParam_T *ga_param)
{
#ifdef DBG
	DBGW(10, ("PrintBestGeneAtPresent enter.\n"));
#endif /* DBG */

	printf("generation[%d], Obtain: %d, Fitness: %e, Gene: (",
		generation, ga_param->bestGene.obtainedGeneration,
		ga_param->bestGene.gaGene.fitness);

	print_ga_gene(&(ga_param->bestGene.gaGene), ga_param, stdout);

	printf(")\n");

#ifdef DBG
	DBGW(10, ("PrintBestGeneAtPresent exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}
	
/*-----
 * PrintFileAllؿ
 *	GaParam_T¤Τǻꤵ줿ϥե
 *	ξΤξϤ
 */
int PrintFileAll(int generation, GaParam_T *ga_param, GaIsland_T *islands)
{
	int	i,j;
	int	num_row;
	int	num_col;
	int	num_var;
	int	rv = EXIT_ERROR;
	FILE	*fp;

#ifdef DBG
	DBGW(10, ("PrintFileAll enter.\n"));
#endif /* DBG */

	if((fp = ga_param->ioParam.outputFile) == NULL){
		log_sys(" ## Error: File not opened.\n");
		goto out;
	}
	
	num_row = ga_param->gaTypeParam.numIsland;
	num_col = ga_param->gaTypeParam.islandPopulation;
	num_var = ga_param->gaEncodeParam.numVariable;
	
	fprintf(fp, "generation,%d\n", generation);
	fprintf(fp, "Best gene:\n");
	fprintf(fp, "Obtained generation, fitness");
	for (i = 0; i < num_var; i++){
		fprintf(fp, ", variable[%d]", i);
	}
	fprintf(fp, "\n");
	fprintf(fp, "%d, ", ga_param->bestGene.obtainedGeneration);
	print_fitness(&(ga_param->bestGene.gaGene), fp);
	fprintf(fp, ", ");
	print_ga_gene(&(ga_param->bestGene.gaGene), ga_param, fp);
	fprintf(fp, "\n");
	
	fprintf(fp, "ID, fitness");
	for (i = 0; i < num_var; i++){
		fprintf(fp, ", variable[%d]", i);
	}
	fprintf(fp, "\n");

	for(i = 0; i < num_row; i++){
		for(j = 0; j < num_col; j++){
			fprintf(fp, "[%d][%d], ", i, j);
			print_fitness(&(islands[i].residents[islands[i].sortedIndex[j]]), fp);
			fprintf(fp, ", ");
			print_ga_gene(&(islands[i].residents[islands[i].sortedIndex[j]])
							, ga_param, fp);
			fprintf(fp, "\n");
		}
	}

	rv = EXIT_NORMAL;
out:

#ifdef DBG
	DBGW(10, ("PrintFileAll exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * PrintResultؿ
 *	õλ塤̤ɽ
 */
int PrintResult(GaParam_T *ga_param)
{
#ifdef DBG
	DBGW(10, ("PrintResult enter.\n"));
#endif /* DBG */

	printf("******Search Result******\n");
	printf("Best gene\n");
	printf("Obtained Generation : %d\n", ga_param->bestGene.obtainedGeneration);
	printf("Fitness             : %e\n", ga_param->bestGene.gaGene.fitness);
	printf("Gene                : (");
	print_ga_gene(&(ga_param->bestGene.gaGene), ga_param, stdout);
	printf(")\n");
	printf("*************************\n");	

#ifdef DBG
	DBGW(10, ("PrintResult exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}

/***********************
 * Static functions    *
 ***********************/

/*-----
 * print_read_summary
 */
static int print_read_summary(FILE *fp, char *file_name, int *read_island_pop, int num_island, int num_read_island, int island_pop){
	int i;

#ifdef DBG
	DBGW(10, ("print_read_summary enter.\n"));
#endif /* DBG */

	fprintf(fp, "\n===Initial Generation File==============\n");
	fprintf(fp, "File name:%s\n", file_name);
	fprintf(fp, "Island, Population, # of read, Random, Redundant\n");
	for(i = 0; i < num_island; i++){
		fprintf(fp, "Island[%d], %d, %d, %d, %d\n",
				i, island_pop, read_island_pop[i],
				read_island_pop[i] > island_pop ? 0 : island_pop - read_island_pop[i],
				read_island_pop[i] > island_pop ? read_island_pop[i] - island_pop : 0);
	}
	if(num_island < num_read_island){
		fprintf(fp, "Number of redundant islands: %d\n", num_read_island - num_island);
	}
	fprintf(fp, "========================================\n");

#ifdef DBG
	DBGW(10, ("print_read_summary exit(0)\n"));
#endif /* DBG */

	return 0;
}

/*-----
 * print_fitness
 */
static int print_fitness(const GaGene_T *ga_gene, FILE *fp)
{
#ifdef DBG
	DBGW(10, ("print_fitness enter.\n"));
#endif /* DBG */

	fprintf(fp, "%e", ga_gene->fitness);

#ifdef DBG
	DBGW(10, ("print_fitness exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}

/*-----
 * print_ga_geneؿ
 * 	ؿǻꤵ줿Τξɽ
 * 	Ϸfitness, varaible[0], ... , variable[num_variable - 1]
 * 	㡧-2.55, 0.34, 0.55, 0.89(̵)
 * 	͡EXIT_NORMAL
 * 	GaGene_T*(I), ѿο(I)ϥȥ꡼FILE*(O), 
 */
static int print_ga_gene(GaGene_T *ga_gene, GaParam_T *ga_param, FILE *fp)
{
	int i;
	int	num_var;
	int	gene_length;
	int	quantum_number;
	int	temp_array[NUM_VARIABLE_MAX];
	int	rv = EXIT_ERROR;
	double	print_array[NUM_VARIABLE_MAX];

#ifdef DBG
	DBGW(10, ("print_ga_gene enter.\n"));
#endif /* DBG */
	
	num_var = ga_param->gaEncodeParam.numVariable;
	gene_length = ga_param->gaEncodeParam.geneLength;
	quantum_number = ga_param->gaEncodeParam.quantumNumber;
	
	switch(ga_param->gaEncodeParam.encodeType){
	case REAL:
		fprintf(fp, "%f", ga_gene->gene_r[0]);
		for(i = 1; i < num_var; i++){
			fprintf(fp, ", %f", ga_gene->gene_r[i]);
		}
		break;
	case BINARY:
		if(BinaryDecode(ga_gene->gene_b, gene_length, quantum_number,
							num_var, temp_array) == EXIT_ERROR){
			goto out;
		}
		Int2Double(&(ga_param->variableRegion), 
								temp_array, print_array, num_var);
		fprintf(fp, "%f", print_array[0]);
		for(i = 1; i < num_var; i++){
			fprintf(fp, ", %f", print_array[i]);
		}
		break;
	case GRAY:
		if(GrayDecode(ga_gene->gene_b, gene_length, quantum_number,
							num_var, temp_array) == EXIT_ERROR){
			goto out;
		}
		Int2Double(&(ga_param->variableRegion), 
								temp_array, print_array, num_var);
		fprintf(fp, "%f", print_array[0]);
		for(i = 1; i < num_var; i++){
			fprintf(fp, ", %f", print_array[i]);
		}
		break;
	}
	
	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("print_ga_gene exit(%d)\n", rv));
#endif /* DBG */
	
	return rv;
}
