#include <math.h>

#include "life.h"
#include "ga2ls.h"
#include "lsInterface.h"
#include "lsUsrFitness.h"
#include "utl_alloc.h"

static int numRow;
static double tStart;
static double tEnd;
static double deltaT;
static double dtSave;
static Parameter defaultParam;
static Parameter defaultParamT;
static Parameter defaultParamP;
static double yOrigin[MAX_VAR_NUM];

static void set_param_default();

int AllocSolutionMatrix(){
	int rv = EXIT_ERROR;
	int num_row;
	int num_col;

#ifdef DBG
	DBGW(10, ("AllocSolutionMatrix enter.\n"));
#endif /* DBG */

	num_row = numRow;
	num_col = Param.n_var + 1;

	if((SolutionMatrix = DoubleMatrixAlloc(num_row, num_col)) == NULL){
		goto out;
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("AllocSolutionMatrix exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

void FreeSolutionMatrix(){
#ifdef DBG
	DBGW(10, ("FreeSolutionMatrix enter.\n"));
#endif /* DBG */
	FreeDoubleMatrix(SolutionMatrix);
#ifdef DBG
	DBGW(10, ("FreeSolutionMatrix exit\n"));
#endif /* DBG */
}

void PreserveDefaultLsParam(double t_start, double t_end, double h, double dt_save, double *y){
#ifdef DBG
	DBGW(10, ("PreserveDefaultLsParam enter.\n"));
#endif /* DBG */

	tStart = t_start;
	tEnd = t_end;
	deltaT = h;
	dtSave = dt_save;
	memcpy(yOrigin, y, MAX_VAR_NUM * sizeof(double));

	numRow = (int)((t_end - t_start)/dt_save) + MAX_EVENT_NUM;

	defaultParam = Param;
	defaultParamT = ParamT;
	defaultParamP = ParamP;
#ifdef DBG
	DBGW(10, ("PreserveDefaultLsParam exit\n"));
#endif /* DBG */

}


int LsFitness(const double *x, int num_variable, int argc, char **argv, double *fitness){
	int	i;
	int	ret, rv;
	int num_row, num_col;
	double y[MAX_VAR_NUM], y_pre[MAX_VAR_NUM], y_start[MAX_VAR_NUM];
	Parameter param_start;

#ifdef DBG
	DBGW(10, ("LsFitness enter.\n"));
#endif /* DBG */

	set_param_default();
	param_start = Param;
	StepNum = 0;
	for(i = 0; i <= Param.n_var; i++) y_start[i] = y_pre[i] = y[i] = yOrigin[i];

	GaGene2LSParam(x);

	ret = DoSolver(param_start, deltaT, tStart, tEnd, dtSave, y_start, y, y_pre);

	num_row = StepNum;
	num_col = Param.n_var;
	
	rv = LsUsrFitness(SolutionMatrix, num_row, num_col, ret, fitness);

#ifdef DBG
	DBGW(10, ("LsFitness exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
		
static void set_param_default()
{
#ifdef DBG
	DBGW(10, ("set_param_default enter.\n"));
#endif /* DBG */

	Param = defaultParam;
	ParamT = defaultParamT;
	ParamP = defaultParamP;

#ifdef DBG
	DBGW(10, ("set_param_default exit\n"));
#endif /* DBG */
}


