#include	<stdio.h>
#include	<string.h>
#include	<math.h>

#include	"logUtil.h"

#include	"matrix.h"
#include	"svd.h"
#include	"c_que.h"

typedef struct	_que_index {
	c_que_t		que;
	m_index_t	index;
} que_index_t;

static int _enque(int i, int j, que_index_t *head);
static int _mtx_norm_matrix_1(m_matrix_t *mtx, double *norm);
static int _mtx_norm_matrix_2(m_matrix_t *mtx, double *norm);
static int _mtx_norm_matrix_inf(m_matrix_t *mtx, double *norm);
static int _mtx_norm_matrix(m_matrix_t *mtx, int p, double *norm);
static int _mtx_norm_vector(m_matrix_t *mtx, int p, double *norm);
static int _mtx_cumprod_vector(m_matrix_t *mtx, int dim, m_matrix_t *cum);
static int _mtx_cumprod_matrix(m_matrix_t *mtx, int dim, m_matrix_t *cum);


/*****************************************************************************
 * ؿ̾
 *      mtx_isdiagonal: гѹĴ٤롣
 * ǽ
 *      гѹĴ٤롣
 *
 *      
 *      1: гǤå
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *      1:	гѹ
 *      0:	гѹ
 * 
 *****************************************************************************/
int mtx_isdiagonal(m_matrix_t *mtx)
{
	int	isdiagonal = 1;
	int	i, j;

	for( i = 0 ; i < mtx->row ; i++ ){
		for( j = 0 ; j < mtx->col ; j++ ){
			if( i == j )	continue;
			if( mtx->elem[i*(mtx->col) + j] == 0 )	continue;
			isdiagonal = 0;
			goto end;
		}
	}
end:
	return isdiagonal;
}
/*****************************************************************************
 * ؿ̾
 *      mtx_issparse: 󤬥ѡĴ٤롣
 * ǽ
 *      󤬥ѡĴ٤롣
 *
 *      
 *      1: 
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *      1:	ѡ
 *      0:	󥹥ѡ
 * 
 *	Not Support
 *****************************************************************************/
int mtx_issparse(m_matrix_t *mtx)
{
	/* NOT Support */
	return 0;
}
/*****************************************************************************
 * ؿ̾
 *      mtx_unit: ñ̹롣
 * ǽ
 *      ñ̹롣
 *
 *      
 *      1: ñ̹롣
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_unit(m_matrix_t *mtx)
{
	int	rowmax, colmax, size, i;
	double	*ptr;

	rowmax = mtx->row;
	colmax = mtx->col;
	if( rowmax != colmax )	goto err;

	size = rowmax * colmax * sizeof(double);
	memset(mtx->elem, 0, size);

	ptr = mtx->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		ptr[i * colmax + i] = 1.0;
	}

	return 0;
err:
	log_sys("ERROR: mtx_unit()\n");
	return -1;
}
/*****************************************************************************
 * ؿ̾
 *      mtx_unique: ǤƱͤι롣
 * ǽ
 *      ǤƱͤι롣
 *
 *      
 *      1: ǤƱͤ롣
 * 
 *      double s;		I: 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_unique(double s, m_matrix_t *mtx)
{
	int	rowmax, colmax, i, j;
	double	*ptr;

	rowmax = mtx->row;
	colmax = mtx->col;

	ptr = mtx->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		for( j = 0 ; j < colmax ; j++ ){
			ptr[i * colmax + j] = s;
		}
	}

	return 0;
}
/*****************************************************************************
 * ؿ̾
 *      mtx_nonzero: 󥼥Ǥο롣
 * ǽ
 *      󥼥Ǥο롣
 *
 *      
 *      1: 󥼥Ǥο롣
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *      󥼥Ǥο
 * 
 *****************************************************************************/
int mtx_nonzero(m_matrix_t *mtx)
{
	int	i, max, cnt;
	double	*ptr;

	cnt = 0;
	max = mtx->row * mtx->col;
	ptr = mtx->elem;
	for( i = 0 ; i < max ; i++ ){
		if( ptr[i] == 0 )	continue;
		cnt++;
	}


	return cnt;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_count: ͤƱǤο롣
 * ǽ
 *      ͤƱǤο롣
 *
 *      
 *      1: ͤƱǤο롣
 * 
 *      double val;		I: Ӥ
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *      פǿ
 * 
 *****************************************************************************/
int mtx_elem_count(double val, m_matrix_t *mtx)
{
	int	i, max, cnt;
	double	*ptr;

	cnt = 0;
	max = mtx->row * mtx->col;
	ptr = mtx->elem;
	for( i = 0 ; i < max ; i++ ){
		if( ptr[i] != val )	continue;
		cnt++;
	}


	return cnt;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_sign: ϹǤǤˤ롣
 * ǽ
 *      ϹǤǤˤ롣
 *
 *      
 *      1: ϹǤǤˤ롣
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 *      m_matrix_t *sgn;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *	ϹƱΥǡǤϰʲΤ줫ͤġ
 *	ϹǤξ硧   1
 *	ϹǤξ硧 0
 *	ϹǤξ硧  -1
 *****************************************************************************/
int mtx_sign(m_matrix_t *mtx, m_matrix_t *sgn)
{
	int	i, j;
	int	rowmax, colmax;
	int	index_r, index;
	double	*m, *s;

	rowmax = mtx->row;
	colmax = mtx->col;

	if( rowmax != sgn->row || colmax != sgn->col ){
		goto err;
	}

	m = mtx->elem;
	s = sgn->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		index_r = i * colmax;
		for( j = 0 ; j < colmax ; j++ ){
			index = index_r + j;
			if( m[index] > 0 ){
				s[index] = 1;
			}else if( m[index] < 0 ){
				s[index] = -1;
			}else{
				s[index] = 0;
			}
		}
	}

	return 0;
err:
	log_sys("ERROR: mtx_sign()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_any: ӾǤ뤫ɤȽꡣ
 * ǽ
 *      ӾǤ뤫ɤȽꡣ
 *
 *      
 *      1: ӾǤ뤫ɤȽꡣ
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 *      int relation;		I: Ӵط
 *      double val;		I: Ӥ
 * ؿ
 *       1:	Ǥͭ
 *       0:	Ǥ̵
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_any(m_matrix_t *mtx, int relation, double val)
{
	int	i, j;
	int	rowmax, colmax;
	double	*ptr;
	int	found = 0;

	rowmax = mtx->row;
	colmax = mtx->col;
	ptr = mtx->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		for( j = 0 ; j < colmax ; j++ ){
			switch( relation ){
			case	MTX_EQ:
				if( ptr[i * colmax + j] == val ){
					found = 1;
					goto end;
				}
				break;
			case	MTX_NOTEQ:
				if( ptr[i * colmax + j] != val ){
					found = 1;
					goto end;
				}
				break;
			case	MTX_LESS:
				if( ptr[i * colmax + j] < val ){
					found = 1;
					goto end;
				}
				break;
			case	MTX_LESSEQ:
				if( ptr[i * colmax + j] <= val ){
					found = 1;
					goto end;
				}
				break;
			case	MTX_MORE:
				if( ptr[i * colmax + j] > val ){
					found = 1;
					goto end;
				}
				break;
			case	MTX_MOREEQ:
				if( ptr[i * colmax + j] >= val ){
					found = 1;
					goto end;
				}
				break;
			default	:
				goto err;
			}
		}
	}

end:
	return found;
err:
	log_sys("ERROR: mtx_any()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_index: ӾǤΥǥåʹԡֹˤ롣
 * ǽ
 *      ӾǤΥǥåʹԡֹˤ롣
 *
 *      
 *      1: ӾǤΥǥå򥭥塼롣
 *      2: ʤ
 * 
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 *      int relation;		I: Ӵط
 *      double val;		I: Ӥ
 *      m_index_t **indexpp;	O: ӾǤΥǥå
 * ؿ
 *      >=0:	ʥǥåο
 *      -1 :	۾
 * 
 *****************************************************************************/
int mtx_index(m_matrix_t *mtx, int relation, double val, m_index_t **indexpp)
{
	int	i, j, len;
	int	rowmax, colmax;
	double	*ptr;
	que_index_t	head, *que;
	m_index_t	*iptr, *index;


	c_initque(&head.que);

	rowmax = mtx->row;
	colmax = mtx->col;
	ptr = mtx->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		for( j = 0 ; j < colmax ; j++ ){
			switch( relation ){
			case	MTX_EQ:
				if( ptr[i * colmax + j] == val ){
					_enque(i, j, &head);
				}
				break;
			case	MTX_NOTEQ:
				if( ptr[i * colmax + j] != val ){
					_enque(i, j, &head);
				}
				break;
			case	MTX_LESS:
				if( ptr[i * colmax + j] < val ){
					_enque(i, j, &head);
				}
				break;
			case	MTX_LESSEQ:
				if( ptr[i * colmax + j] <= val ){
					_enque(i, j, &head);
				}
				break;
			case	MTX_MORE:
				if( ptr[i * colmax + j] > val ){
					_enque(i, j, &head);
				}
				break;
			case	MTX_MOREEQ:
				if( ptr[i * colmax + j] >= val ){
					_enque(i, j, &head);
				}
				break;
			default	:
				goto err;
			}
		}
	}


	index = NULL;
	len = c_quelen(&head.que);
	if( len == 0 ){
		goto end;
	}
	index = (m_index_t *) malloc(len * sizeof(m_index_t));
	if( index == NULL ){
		goto err;
	}

	iptr = index;
	for( i = len - 1 ; i >= 0 ; i-- ){
		que = (que_index_t *) c_deque(&head.que);
		memcpy(iptr, &que->index, sizeof(m_index_t));
		iptr++;
		free(que);
	}

end:
	*indexpp = index;
	return len;
err:
	len = c_quelen(&head.que);
	for( i = len - 1 ; i <= 0 ; i-- ){
		que = (que_index_t *) c_deque(&head.que);
		free(que);
	}
	*indexpp = NULL;
	log_sys("ERROR: mtx_index()\n");
	return -1;
}
/*
 * 塼Ͽ롣
 */
static int _enque(int i, int j, que_index_t *head)
{
	que_index_t	*que;

	que = (que_index_t *) malloc(sizeof(que_index_t));
	if( que == NULL ){
		goto err;
	}
	memset(que, 0, sizeof(que_index_t));

	que->index.i = i;
	que->index.j = j;

	c_enque(&head->que, &que->que);

	return 0;
err:
	log_sys("ERROR: _enque()\n");
	return -1;
}
/*****************************************************************************
 * ؿ̾
 *      mtx_index_free: mtx_index() Ǽǥåγ
 * ǽ
 *      mtx_index() ǼǥåΥ꥽롣
 *
 *      
 *      1: mtx_index() ǼǥåΥ꥽롣
 * 
 *      m_index_t *index;	I: ӾǤΥǥå
 * ؿ
 *      ̵
 * 
 *****************************************************************************/
void mtx_index_free(m_index_t *index)
{
	if( index )	free(index);
}
/*****************************************************************************
 * ؿ̾
 *      mtx_set_elem: mtx_index() ǼǥåǤͤꡣ
 * ǽ
 *      mtx_index() ǼǥåǤͤꤹ롣
 *
 *      
 *      1: mtx_index() ǼǥåǤͤꤹ롣
 * 
 *      double val;		I: 
 *      int len;		I: ǥåο
 *      m_index_t *index;	I: ӾǤΥǥå
 *      m_matrix_t *mtx;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_set_elem(double val, int len, m_index_t *index, m_matrix_t *mtx)
{
	int	k, colmax;
	double	*ptr;

	colmax = mtx->col;
	ptr = mtx->elem;
	for( k = 0 ; k < len ; k++ ){
		ptr[index[k].i * colmax + index[k].j] = val;
	}

	return 0;
}


/*****************************************************************************
 * ؿ̾
 *      mtx_addup: 
 * ǽ
 *      ¤Ȥ롣
 *
 *      
 *      1: ¤Ȥ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MB;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_addup(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *b, *x;
	int	arow, acol;
	int	brow, bcol;
	int	xrow, xcol;

	a = MA->elem;
	b = MB->elem;
	x = MX->elem;
	if( a == NULL || b == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	brow = MB->row;
	bcol = MB->col;
	if( brow <= 0 || bcol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* ·׻ǽ */
	if( arow != brow || acol != bcol ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = *a + *b;
		x++;
		a++;
		b++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_addup()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_subtract: κ
 * ǽ
 *      κȤ롣
 *
 *      
 *      1: κȤ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MB;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_subtract(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *b, *x;
	int	arow, acol;
	int	brow, bcol;
	int	xrow, xcol;

	a = MA->elem;
	b = MB->elem;
	x = MX->elem;
	if( a == NULL || b == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	brow = MB->row;
	bcol = MB->col;
	if( brow <= 0 || bcol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* ·׻ǽ */
	if( arow != brow || acol != bcol ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = *a - *b;
		x++;
		a++;
		b++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_subtract()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_transpose: žֹ롣
 * ǽ
 *      žֹ롣
 *
 *      
 *      1: žֹ롣
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *trg;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_transpose(m_matrix_t *src, m_matrix_t *trg)
{
	int	i, j;
	int	rowmax, colmax;
	double	*s_ptr, *t_ptr;

	rowmax = trg->row;
	colmax = trg->col;
	s_ptr = src->elem;
	t_ptr = trg->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		for( j = 0 ; j < colmax ; j++ ){
			t_ptr[i * colmax + j] = s_ptr[j * rowmax + i];
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_multiply: ѤȤ롣
 * ǽ
 *      ѤȤ롣
 *
 *      
 *      1: ѤȤ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MB;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_multiply(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX)
{
	int	i, j, k, kk;
	double	*a, *b, *x;
	double	*ai, *bj, *xj;
	double	s;
	int	arow, acol;
	int	brow, bcol;
	int	xrow, xcol;

	a = MA->elem;
	b = MB->elem;
	x = MX->elem;
	if( a == NULL || b == NULL || x == NULL ){
		log_sys("mtx_multiply(%p, %p, %p)\n", a, b, x);
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		log_sys("A[%d, %d]\n", arow, acol);
		goto err;
	}
	brow = MB->row;
	bcol = MB->col;
	if( brow <= 0 || bcol <= 0 ){
		log_sys("B[%d, %d]\n", brow, bcol);
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		log_sys( "X[%d, %d]\n", xrow, xcol);
		goto err;
	}
	/* ѷ׻ǽ */
	if( acol != brow ){
		log_sys("A-col[%d] != brow[%d]\n", acol, brow);
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != bcol ){
		log_sys("X-row[%d] != A-row[%d], X-col[%d] != bcol[%d]\n",
				xrow, arow, xcol, bcol);
		goto err;
	}

	for( j = 0, bj = b, xj = x ; j < bcol ; j++, bj++, xj++ ){
		for( i = 0, ai = a ; i < arow ; i++, ai += acol ){
			s = 0.0;
			for( k = 0, kk = 0 ; k < acol ; k++, kk += bcol ){
				s += ai[k] * bj[kk];
			}
			xj[i * xcol] = s;
		}
	}

	return 0;
err:
	log_sys("ERROR: mtx_multiply()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_multiply: ƱΤѤȤ롣
 * ǽ
 *      ƱΤѤȤ롣
 *
 *      
 *      1: ƱΤѤȤ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MB;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_multiply(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *b, *x;
	int	arow, acol;
	int	brow, bcol;
	int	xrow, xcol;

	a = MA->elem;
	b = MB->elem;
	x = MX->elem;
	if( a == NULL || b == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	brow = MB->row;
	bcol = MB->col;
	if( brow <= 0 || bcol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* ѷ׻ǽ */
	if( arow != brow || acol != bcol ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = (*a) * (*b);
		x++;
		a++;
		b++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_multiply()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_multiply: ƱΤξȤ롣
 * ǽ
 *      ƱΤξȤ롣
 *
 *      
 *      1: ƱΤξȤ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MB;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_divide(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *b, *x;
	int	arow, acol;
	int	brow, bcol;
	int	xrow, xcol;

	a = MA->elem;
	b = MB->elem;
	x = MX->elem;
	if( a == NULL || b == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	brow = MB->row;
	bcol = MB->col;
	if( brow <= 0 || bcol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* ѷ׻ǽ */
	if( arow != brow || acol != bcol ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = (*a) / (*b);
		x++;
		a++;
		b++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_divide()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_multiply: ƱΤΤ٤Ȥ롣
 * ǽ
 *      ƱΤΤ٤(mx = ma^mb)Ȥ롣
 *
 *      
 *      1: ƱΤΤ٤Ȥ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MB;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_power(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *b, *x;
	int	arow, acol;
	int	brow, bcol;
	int	xrow, xcol;

	a = MA->elem;
	b = MB->elem;
	x = MX->elem;
	if( a == NULL || b == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	brow = MB->row;
	bcol = MB->col;
	if( brow <= 0 || bcol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* ѷ׻ǽ */
	if( arow != brow || acol != bcol ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = pow((*a), (*b));
		x++;
		a++;
		b++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_power()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_addup_to_scalar: 顼ȹ¡
 * ǽ
 *      Ǥ˥顼̤ä롣
 *
 *      
 *      1: Ǥ˥顼̤ä롣
 * 
 *      double p;	I: 顼
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_addup_to_scalar(double p, m_matrix_t *MA, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = p + *a;
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_addup_to_scalar()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_subtract_from_scalar: 顼ȹκ
 * ǽ
 *      顼̤γǤͤǤ˻Ĺ롣
 *
 *      
 *      1: 顼̤γǤͤǤ˻Ĺ롣
 * 
 *      double p;	I: 顼
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_subtract_from_scalar(double p, m_matrix_t *MA, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		log_sys("mtx_elem_subtract_from_scalar(NULL)\n");
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		log_sys("mtx_elem_subtract_from_scalar(a.r/c<=0)\n");
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		log_sys("mtx_elem_subtract_from_scalar(x.r/c<=0)\n");
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		log_sys("mtx_elem_subtract_from_scalar(x.r/c!=a.r/c)\n");
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = p - *a;
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_subtract_from_scalar()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_multiply_by_scalar: 顼ȹѡ
 * ǽ
 *      顼̤ȹγǤݤͤǤ˻Ĺ롣
 *
 *      
 *      1: 顼̤ȹγǤݤͤǤ˻Ĺ롣
 * 
 *      double p;	I: 顼
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_multiply_by_scalar(double p, m_matrix_t *MA, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = p * (*a);
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_multiply_by_scalar()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_elem_divide_into_scalar: 顼ȹξ
 * ǽ
 *      顼̤γǤǳäͤǤ˻Ĺ롣
 *
 *      
 *      1: 顼̤γǤǳäͤǤ˻Ĺ롣
 * 
 *      double p;	I: 顼
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_elem_divide_into_scalar(double p, m_matrix_t *MA, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = p / *a;
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_elem_divide_into_scalar()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_scalar_addup: ȥ顼¡
 * ǽ
 *      γǤ˥顼̤äͤǤ˻Ĺ롣
 *
 *      
 *      1: γǤ˥顼̤äͤǤ˻Ĺ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      double p;	I: 顼
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_scalar_addup(m_matrix_t *MA, double p, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = *a + p;
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_scalar_addup()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_scalar_multiply: ȥ顼ѡ
 * ǽ
 *      γǤ˥顼̤ݤͤǤ˻Ĺ롣
 *
 *      
 *      1: γǤ˥顼̤ݤͤǤ˻Ĺ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      double p;	I: 顼
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_scalar_multiply(m_matrix_t *MA, double p, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = (*a) * p;
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_scalar_multiply()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_scalar_addup: ȥ顼Τ٤衣
 * ǽ
 *      γǤ򥹥顼̤Τ٤褷ͤǤ˻Ĺ롣
 *
 *      
 *      1: γǤ򥹥顼̤Τ٤褷ͤǤ˻Ĺ롣
 * 
 *      m_matrix_t *MA;	I: ¤ΤؤΥݥ
 *      double p;	I: 顼
 *      m_matrix_t *MX;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_scalar_power(m_matrix_t *MA, double p, m_matrix_t *MX)
{
	int	i, max;
	double	*a, *x;
	int	arow, acol;
	int	xrow, xcol;

	a = MA->elem;
	x = MX->elem;
	if( a == NULL || x == NULL ){
		goto err;
	}
	arow = MA->row;
	acol = MA->col;
	if( arow <= 0 || acol <= 0 ){
		goto err;
	}
	xrow = MX->row;
	xcol = MX->col;
	if( xrow <= 0 || xcol <= 0 ){
		goto err;
	}
	/* Ǽꥢ­ */
	if( xrow != arow || xcol != acol ){
		goto err;
	}

	max = xrow * xcol;
	for( i = 0 ; i < max ; i++ ){
		*x = pow((*a), p);
		x++;
		a++;
	}

	return 0;
err:
	log_sys("ERROR: mtx_scalar_power()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_diagonal: ٥ȥгѹѴ롣
 * ǽ
 *      ٥ȥǤгǤ˻гѹ롣
 *
 *      
 *      1: ٥ȥǤгǤ˻гѹ롣
 * 
 *      m_matrix_t *vec;	I: ¤ΤؤΥݥ
 *      m_matrix_t *diag;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *	гѹгǤϡΥ٥ȥǤ
 *****************************************************************************/
int mtx_diagonal(m_matrix_t *vec, m_matrix_t *diag)
{
	int	max, colmax, size, i;
	double	*vec_ptr, *mtx_ptr;

	if( vec->row <= 0 || vec->col <= 0 ){
		goto err;
	}
	if( diag->row <= 0 || diag->col <= 0 ){
		goto err;
	}

	if( vec->row == 1 ){
		max = vec->col;
	}else{
		if( vec->col != 1 ){
			goto err;
		}
		max = vec->row;
	}

	size = diag->row * diag->col * sizeof(double);
	memset(diag->elem, 0, size);
	colmax = diag->col;
	vec_ptr = vec->elem;
	mtx_ptr = diag->elem;
	for( i = 0 ; i < max ; i++ ){
		mtx_ptr[i * colmax + i] = vec_ptr[i];
	}

	return 0;
err:
	log_sys("ERROR: mtx_diagonal()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_diagonal2vector: гǤ٥ȥѴ롣
 * ǽ
 *      гǤ٥ȥѴ롣
 *
 *      
 *      1: гǤ٥ȥѴ롣
 * 
 *      int k;			I: гѤεΥ
 *      m_matrix_t *diag;	I: ¤ΤؤΥݥ
 *      m_matrix_t *vec;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 * 	k == 0: гѡ
 * 	k > 0 :	гѤα塣
 * 	k < 0 :	гѤκ
 *****************************************************************************/
int mtx_diagonal2vector(int k, m_matrix_t *diag, m_matrix_t *vec)
{
	int	max, v_max, colmax, i;
	double	*vec_ptr, *mtx_ptr;

	if( vec->row <= 0 || vec->col <= 0 ){
		goto err;
	}
	if( diag->row <= 0 || diag->col <= 0 ){
		goto err;
	}

	colmax = diag->col;
	max = colmax - abs(k);
	if( vec->row == 1 ){
		v_max = vec->col;
	}else{
		if( vec->col != 1 ){
			goto err;
		}
		v_max = vec->row;
	}
	if( v_max < max ){
		goto err;
	}
	if( v_max > max ){
		if( vec->row == 1 ){
			vec->col = max;
		}else{
			vec->row = max;
		}
	}

	vec_ptr = vec->elem;
	mtx_ptr = diag->elem;
	for( i = 0 ; i < max ; i++ ){
		if( k >= 0 ){
			vec_ptr[i] = mtx_ptr[i * colmax + (i + k)];
		}else{
			vec_ptr[i] = mtx_ptr[(i - k) * colmax + i];
		}
	}

	return 0;
err:
	log_sys("ERROR: mtx_diagonal2vector()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_sum: ٥ȥǤ¤Ǻäԥ٥ȥ֤
 * ǽ
 *      ٥ȥǤ¤Ǻäԥ٥ȥ֤
 *	٥ȥξ硧	Ǥ
 *	ξ硧		٥ȥ¤Ǻäԥ٥ȥ롣
 *
 *      
 *      1: ٥ȥǤ¤Ǻäԥ٥ȥ֤
 * 
 *      int sumdir;		I: ¤Ȥ󡦹ԡ
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 *      m_matrix_t *sum;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_sum(int sumdir, m_matrix_t *mtx, m_matrix_t *sum)
{
	int	i, j, rowmax, colmax;
	double	s;
	double	*ptr;

	if( mtx->row <= 0 || mtx->col <= 0 ){
		log_sys("ERROR: mtx_sum()\n");
		return -1;
	}

#if 0
	if( mtx->row == 1 || mtx->col == 1 ){
	/* ٥ȥǤ */
		max = mtx->row * mtx->col;
		s = 0.;
		ptr = mtx->elem;
		for( i = 0 ; i < max ; i++ ){
			s += ptr[i];
		}
		sum->elem[0] = s;
		return 0;
	}
#endif

	/*  */
	rowmax = mtx->row;
	colmax = mtx->col;
	ptr = mtx->elem;

	if( sumdir == MTX_COL ){
	/* ٥ȥǤ¤ */
		for( j = 0 ; j < colmax ; j++ ){
			s = 0.;
			for( i = 0 ; i < rowmax ; i++ ){
				s += ptr[i*colmax + j];
			}
			sum->elem[j] = s;
		}
	}else{
	/* ιԥ٥ȥǤ¤ */
		for( i = 0 ; i < rowmax ; i++ ){
			s = 0.;
			for( j = 0 ; j < colmax ; j++ ){
				s += ptr[i*colmax + j];
			}
			sum->elem[i] = s;
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_abs: Ǥ֤ͤ롣
 * ǽ
 *      Ǥ֤ͤ롣
 *
 *      
 *      1: Ǥ֤ͤ롣
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *abs;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_abs(m_matrix_t *src, m_matrix_t *abs)
{
	int	i, max;
	double	*src_ptr, *abs_ptr;

	if( src->row <= 0 || src->col <= 0 ){
		log_sys("ERROR: mtx_abs()\n");
		return -1;
	}

	max = src->row * src->col;
	src_ptr = src->elem;
	abs_ptr = abs->elem;
	for( i = 0 ; i < max ; i++ ){
		*abs_ptr = fabs(*src_ptr);
		src_ptr++;
		abs_ptr++;
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_min: ǾǤޤ٥ȥ֤
 * ǽ
 *      ǾǤޤ٥ȥ֤
 *	٥ȥξ硧	Ǿ
 *	ξ硧		٥ȥκǾǤǺäԥ٥ȥ롣
 *
 *      
 *      1: ǾǤޤ٥ȥ֤
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *min;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_min(m_matrix_t *src, m_matrix_t *min)
{
	int	i, j, rowmax, colmax;
	double	*src_ptr, *min_ptr;

	if( src->row <= 0 || src->col <= 0 ){
		log_sys("ERROR: mtx_min()\n");
		return -1;
	}

	rowmax = src->row;
	colmax = src->col;
	src_ptr = src->elem;
	min_ptr = min->elem;
	for( j = 0 ; j < colmax ; j++ ){
		min_ptr[j] = src_ptr[0 * colmax + j];
		for( i = 1 ; i < rowmax ; i++ ){
			if( min_ptr[j] > src_ptr[i * colmax + j] ){
				min_ptr[j] = src_ptr[i * colmax + j];
			}
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_max: Ǥޤ٥ȥ֤
 * ǽ
 *      Ǥޤ٥ȥ֤
 *	٥ȥξ硧	
 *	ξ硧		٥ȥκǤǺäԥ٥ȥ롣
 *
 *      
 *      1: Ǥޤ٥ȥ֤
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *max;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *	[ a0 b0 ... z0 ]
 *	[ a1 b1 ... z1 ]
 *	[       ...    ]
 *	[ ai bi ... zi ]
 *	[       ...    ]
 *	[ an bn ... zn ]
 *
 *	[ max(ai) max(bi) ... max(zi) ]
 *****************************************************************************/
int mtx_max(m_matrix_t *src, m_matrix_t *max)
{
	int	i, j, rowmax, colmax;
	double	*src_ptr, *max_ptr;

	if( src->row <= 0 || src->col <= 0 ){
		log_sys("ERROR: mtx_max()\n");
		return -1;
	}

	rowmax = src->row;
	colmax = src->col;
	src_ptr = src->elem;
	max_ptr = max->elem;
	for( j = 0 ; j < colmax ; j++ ){
		max_ptr[j] = src_ptr[0 * colmax + j];
		for( i = 1 ; i < rowmax ; i++ ){
			if( max_ptr[j] < src_ptr[i * colmax + j] ){
				max_ptr[j] = src_ptr[i * colmax + j];
			}
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_max_index: Ǥޤ٥ȥ֤
 * ǽ
 *      Ǥޤ٥ȥ֤
 *	٥ȥξ硧	
 *	ξ硧		٥ȥκǤǺäԥ٥ȥ롣
 *	ǤΥǥå֤
 *
 *      
 *      1: Ǥޤ٥ȥ֤
 *	2: ǤΥǥå֤
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *max;	O: ¤ΤؤΥݥ
 *      int index[];		O: ǥåؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *	[ a0 b0 ... z0 ]
 *	[ a1 b1 ... z1 ]
 *	[       ...    ]
 *	[ ai bi ... zi ]
 *	[       ...    ]
 *	[ an bn ... zn ]
 *
 *	[ max(ai) max(bi) ... max(zi) ]	-> κǤ
 *	[ i_max_a i_max_b ... i_max_z ]	-> κǤΥǥå
 *****************************************************************************/
int mtx_max_index(m_matrix_t *src, m_matrix_t *max, int index[])
{
	int	i, j, rowmax, colmax;
	double	*src_ptr, *max_ptr;

	if( src->row <= 0 || src->col <= 0 ){
		log_sys("ERROR: mtx_max()\n");
		return -1;
	}

	rowmax = src->row;
	colmax = src->col;
	src_ptr = src->elem;
	max_ptr = max->elem;
	for( j = 0 ; j < colmax ; j++ ){
		max_ptr[j] = src_ptr[0 * colmax + j];
		index[j] = 0;
		for( i = 1 ; i < rowmax ; i++ ){
			if( max_ptr[j] < src_ptr[i * colmax + j] ){
				max_ptr[j] = src_ptr[i * colmax + j];
				index[j] = i;
			}
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_max_row: Ǥޤ٥ȥ֤
 * ǽ
 *      Ǥޤ٥ȥ֤
 *	٥ȥξ硧	
 *	ξ硧		ƹԥ٥ȥκǤǺä٥ȥ롣
 *
 *      
 *      1: Ǥޤ٥ȥ֤
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *max;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *	[ a1 a2 ... ai ... an]		[ max(ai) ]
 *	[ b1 b2 ... bi ... bn]		[ max(bi) ]
 *	[       ...          ] 	  ->	[   ...   ]
 *	[       ...          ]		[   ...   ]
 *	[ z1 z2 ... zi ... zn]		[ max(zi) ]
 *****************************************************************************/
int mtx_max_row(m_matrix_t *src, m_matrix_t *max)
{
	int	i, j, rowmax, colmax;
	double	*src_ptr, *max_ptr;

	if( src->row <= 0 || src->col <= 0 ){
		log_sys("ERROR: mtx_max_row()\n");
		return -1;
	}

	rowmax = src->row;
	colmax = src->col;
	src_ptr = src->elem;
	max_ptr = max->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		max_ptr[i] = src_ptr[i * colmax + 0];
		for( j = 1 ; j < colmax ; j++ ){
			if( max_ptr[i] < src_ptr[i * colmax + j] ){
				max_ptr[i] = src_ptr[i * colmax + j];
			}
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_max2: ĤιǤӤ礭ǤȤ
 * ǽ
 *	Ĥιʥ٥ȥˤƱǥåǤӤơ
 *	礭ǤˤƱƱιʥ٥ȥˤ֤
 *
 *      
 *      1: Ĥιʥ٥ȥˤƱǥåǤӡ
 *	2: 礭ǤˤƱƱιʥ٥ȥˤꡣ
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *max;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_max2(m_matrix_t *src0, m_matrix_t *src1, m_matrix_t *max)
{
	int	i, j, rowmax, colmax, index_r, index;
	double	*src0_ptr, *src1_ptr, *max_ptr;

	if( src0->row <= 0 || src0->col <= 0 ){
		log_sys("ERROR: mtx_max2() src0[%d, %d]\n",
					src0->row, src0->col);
		return -1;
	}
	if( src1->row <= 0 || src1->col <= 0 ){
		log_sys("ERROR: mtx_max2() src1[%d, %d]\n",
					src1->row, src1->col);
		return -1;
	}

	rowmax = src0->row;
	colmax = src0->col;
	src0_ptr = src0->elem;
	src1_ptr = src1->elem;
	max_ptr = max->elem;
	for( i = 0 ; i < rowmax ; i++ ){
		index_r = i * colmax;
		for( j = 0 ; j < colmax ; j++ ){
			index = index_r + j;
			if( src0_ptr[index] > src1_ptr[index] ){
				max_ptr[index] = src0_ptr[index];
			}else{
				max_ptr[index] = src1_ptr[index];
			}
		}
	}

	return 0;
}

/*
 * Ѥ롣
 */
/*****************************************************************************
 * ؿ̾
 *      mtx_cumprod: Ѥ롣
 * ǽ
 *      ǤѤ롣
 *
 *      
 *      1: Ѥ롣
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *	int dim;		I: Ѥʹԡ
 *      m_matrix_t *cum;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_cumprod(m_matrix_t *mtx, int dim, m_matrix_t *cum)
{
	if( mtx->row <= 0 || mtx->col <= 0 ){
		log_sys("ERROR: mtx_cumprod()\n");
		return -1;
	}

	/* ٥ȥ */
	if( mtx->row == 1 || mtx->col == 1 ){
		return _mtx_cumprod_vector(mtx, dim, cum);
	}

	/*  */
	return _mtx_cumprod_matrix(mtx, dim, cum);
}
static int _mtx_cumprod_vector(m_matrix_t *mtx, int dim, m_matrix_t *cum)
{
	int	max, i;
	double	*src_ptr, *cum_ptr;

	max = ( mtx->row > mtx->col )? mtx->row : mtx->col;
	src_ptr = mtx->elem;
	cum_ptr = cum->elem;
	cum_ptr[0] = src_ptr[0];
	for( i = 1 ; i < max ; i++ ){
		cum_ptr[i] = cum_ptr[i - 1] * src_ptr[i];
	}

	return 0;
}
static int _mtx_cumprod_matrix(m_matrix_t *mtx, int dim, m_matrix_t *cum)
{
	int	i, j, maxrow, maxcol;
	double	*src_ptr, *cum_ptr;

	src_ptr = mtx->elem;
	cum_ptr = cum->elem;
	maxrow = mtx->row;
	maxcol = mtx->col;
	if( dim == 1 ){
		for( j = 0 ; j < maxcol ; j++ ){
			cum_ptr[0 * maxcol + j] = src_ptr[0 * maxcol + j];
			for( i = 1 ; i < maxrow ; i++ ){
				cum_ptr[i * maxcol + j] =
					src_ptr[i * maxcol + j]
					* cum_ptr[(i - 1) * maxcol + j];
			}
		}
	}else{
		for( i = 0 ; i < maxrow ; i++ ){
			cum_ptr[i * maxcol + 0] = src_ptr[i * maxcol + 0];
			for( j = 1 ; j < maxcol ; j++ ){
				cum_ptr[i * maxcol + j] =
					src_ptr[i * maxcol + j]
					* cum_ptr[i * maxcol + j - 1];
			}
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_norm: Υ롣
 * ǽ
 *      ΥΥ롣
 *
 *      
 *      1: ΥΥ롣
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *	int p;			I: Υ(1,2,...;INF;-INF)
 *      m_matrix_t *min;	O: ¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_norm(m_matrix_t *mtx, int p, double *norm)
{
	if( mtx->row <= 0 || mtx->col <= 0 ){
		log_sys("ERROR: mtx_norm()\n");
		return -1;
	}

	/* ٥ȥΥΥ */
	if( mtx->row == 1 || mtx->col == 1 ){
		return _mtx_norm_vector(mtx, p, norm);
	}

	/* ΥΥ */
	return _mtx_norm_matrix(mtx, p, norm);
}
static int _mtx_norm_vector(m_matrix_t *mtx, int p, double *norm)
{
	m_matrix_t	abs;
	m_matrix_t	sum;
	double		sumelem[1];
	double		p_dbl, p_inv;

	memset(&abs, 0, sizeof(m_matrix_t));
	MTX_SET_ALLOC(abs, mtx->row, mtx->col);
	if( abs.elem == NULL ){
		goto err;
	}

	if( mtx_abs(mtx, &abs) < 0 ){
		goto err;
	}

	MTX_SET(sum, 1, 1, sumelem);
	sumelem[0] = 0;

	if( p == PLUS_INF ){
	/* max(abs(V)) */
		if( mtx_max(&abs, &sum) < 0 ){
			goto err;
		}
	}else if( p == MINUS_INF ){
	/* min(abs(V)) */
		if( mtx_min(&abs, &sum) < 0 ){
			goto err;
		}
	}else{
	/* sum(abs(V).^p)^(1/p) */
		if( p < 1 ){
			goto err;
		}
		p_dbl = (double) p;
		p_inv = 1. / p_dbl;
		if( mtx_scalar_power(&abs, p_dbl, &abs) < 0 ){
			goto err;
		}
		if( mtx_sum(MTX_COL, &abs, &sum) < 0 ){
			goto err;
		}
		sum.elem[0] = pow(sum.elem[0], p_inv);
	}
	*norm = sum.elem[0];

	MTX_SET_FREE(abs);
	return 0;
err:
	MTX_SET_FREE(abs);
	log_sys("ERROR: _mtx_norm_vector()\n");
	return -1;
}
static int _mtx_norm_matrix(m_matrix_t *mtx, int p, double *norm)
{
	if( p == 1 ){
	/* max(sum(abs(A))) */
		return _mtx_norm_matrix_1(mtx, norm);
	}else if( p == 2 ){
	/* max(svd(A)) */
		return _mtx_norm_matrix_2(mtx, norm);
	}else if( p == PLUS_INF ){
	/* max(sum(abs(A'))) */
		return _mtx_norm_matrix_inf(mtx, norm);
	}else{
		log_sys("ERROR: _mtx_norm_matrix()\n");
		return -1;
	}
}
static int _mtx_norm_matrix_1(m_matrix_t *mtx, double *norm)
{
	m_matrix_t	abs;
	m_matrix_t	sum;
	m_matrix_t	max;
	double		maxelem[1];

	memset(&abs, 0, sizeof(m_matrix_t));
	memset(&sum, 0, sizeof(m_matrix_t));
	memset(&max, 0, sizeof(m_matrix_t));

	MTX_SET_ALLOC(abs, mtx->row, mtx->col);
	MTX_SET_ALLOC(sum, mtx->row, 1);
	if( abs.elem == NULL || sum.elem == NULL ){
		goto err;
	}
	MTX_SET(max, 1, 1, maxelem);
	maxelem[0] = 0;

	/* max(sum(abs(A))) */
	if( mtx_abs(mtx, &abs) < 0 ){
		goto err;
	}
	if( mtx_sum(MTX_COL, &abs, &sum) < 0 ){
		goto err;
	}
	if( mtx_max(&sum, &max) < 0 ){
		goto err;
	}
	*norm = max.elem[0];

	MTX_SET_FREE(abs);
	MTX_SET_FREE(sum);
	return 0;
err:
	MTX_SET_FREE(abs);
	MTX_SET_FREE(sum);
	log_sys("ERROR: _mtx_norm_matrix_1()\n");
	return -1;
}
static int _mtx_norm_matrix_2(m_matrix_t *mtx, double *norm)
{
	m_matrix_t	m_u, v_w, m_v, max;
	double		maxelem[1];

	memset(&m_u, 0, sizeof(m_matrix_t));
	memset(&v_w, 0, sizeof(m_matrix_t));
	memset(&m_v, 0, sizeof(m_matrix_t));
	memset(&max, 0, sizeof(m_matrix_t));

	MTX_SET_ALLOC(m_u, mtx->row, mtx->col);
	MTX_SET_ALLOC(v_w, mtx->row, 1);
	MTX_SET_ALLOC(m_v, mtx->row, mtx->col);
	if( m_u.elem == NULL || v_w.elem == NULL || m_v.elem == NULL ){
		goto err;
	}
	MTX_SET(max, 1, 1, maxelem);
	maxelem[0] = 0;

	/* max(svd(A)) */
	if( mtx_svdecomp(mtx, &m_u, &v_w, &m_v) < 0 ){
		goto err;
	}
	if( mtx_max(&v_w, &max) < 0 ){
		goto err;
	}
	*norm = max.elem[0];

	MTX_SET_FREE(m_u);
	MTX_SET_FREE(v_w);
	MTX_SET_FREE(m_v);
	return 0;
err:
	MTX_SET_FREE(m_u);
	MTX_SET_FREE(v_w);
	MTX_SET_FREE(m_v);
	log_sys("ERROR: _mtx_norm_matrix_2()\n");
	return -1;
}
static int _mtx_norm_matrix_inf(m_matrix_t *mtx, double *norm)
{
	m_matrix_t	trans;
	m_matrix_t	abs;
	m_matrix_t	sum;
	m_matrix_t	max;
	double		maxelem[1];

	memset(&trans, 0, sizeof(m_matrix_t));
	memset(&abs, 0, sizeof(m_matrix_t));
	memset(&sum, 0, sizeof(m_matrix_t));
	memset(&max, 0, sizeof(m_matrix_t));

	MTX_SET_ALLOC(trans, mtx->col, mtx->row);
	MTX_SET_ALLOC(abs, mtx->col, mtx->row);
	MTX_SET_ALLOC(sum, mtx->col, 1);
	if( trans.elem == NULL || abs.elem == NULL || sum.elem == NULL ){
		goto err;
	}
	MTX_SET(max, 1, 1, maxelem);
	maxelem[0] = 0;

	/* max(sum(abs(A'))) */
	if( mtx_transpose(mtx, &trans) < 0 ){
		goto err;
	}
	if( mtx_abs(&trans, &abs) < 0 ){
		goto err;
	}
	if( mtx_sum(MTX_COL, &abs, &sum) < 0 ){
		goto err;
	}
	if( mtx_max(&sum, &max) < 0 ){
		goto err;
	}
	*norm = max.elem[0];

	MTX_SET_FREE(trans);
	MTX_SET_FREE(abs);
	MTX_SET_FREE(sum);
	return 0;
err:
	MTX_SET_FREE(trans);
	MTX_SET_FREE(abs);
	MTX_SET_FREE(sum);
	log_sys("ERROR: _mtx_norm_matrix_inf()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_get_submatrix: 󤫤ʬФ
 * ǽ
 *      󤫤ʬФ
 *
 *      
 *      1: 󤫤ʬФ
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *trg;	O: ¤ΤؤΥݥ
 *      int from_row;		I: Ϲֹ
 *      int from_col;		I: ֹ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *          +------------------------+
 *          |                        |
 * from_row + --- +------+           | -----> +------+
 *          |     |      |           |        |      |
 *          |     +------+           | -----> +------+
 *          |     |                  |
 *          +-----+------------------+
 *                from_col
 *****************************************************************************/
int mtx_get_submatrix(m_matrix_t *src, m_matrix_t *trg, int from_row, int from_col)
{
	double	*src_ptr, *trg_ptr;
	int	src_rowmax, src_colmax, src_index;
	int	trg_rowmax, trg_colmax, trg_index;
	int	i, j;

	src_ptr = src->elem;
	trg_ptr = trg->elem;
	src_rowmax = src->row;
	src_colmax = src->col;
	trg_rowmax = trg->row;
	trg_colmax = trg->col;

	for( i = 0 ; i < trg_rowmax ; i++ ){
		src_index = (i + from_row) * src_colmax;
		trg_index = i * trg_colmax;
		for( j = 0 ; j < trg_colmax ; j++ ){
			trg_ptr[trg_index + j]
				= src_ptr[src_index + j + from_col];
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_set_submatrix: ΰʬꤹ롣
 * ǽ
 *	ΰʬꤹ롣
 *
 *      
 *      1: ΰʬꤹ롣
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *trg;	O: ¤ΤؤΥݥ
 *      int from_row;		I: Ϲֹ
 *      int from_col;		I: ֹ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *          +------------------------+
 *          |                        |
 * from_row + --- +------+           | <----- +------+
 *          |     |      |           |        |      |
 *          |     +------+           | <----- +------+
 *          |     |                  |
 *          +-----+------------------+
 *                from_col
 *****************************************************************************/
int mtx_set_submatrix(m_matrix_t *src, m_matrix_t *trg, int from_row, int from_col)
{
	double	*src_ptr, *trg_ptr;
	int	src_rowmax, src_colmax, src_index;
	int	trg_rowmax, trg_colmax, trg_index;
	int	i, j;

	src_ptr = src->elem;
	trg_ptr = trg->elem;
	src_rowmax = src->row;
	src_colmax = src->col;
	trg_rowmax = trg->row;
	trg_colmax = trg->col;

	for( i = 0 ; i < src_rowmax ; i++ ){
		src_index = i * src_colmax;
		trg_index = (i + from_row) * trg_colmax;
		for( j = 0 ; j < src_colmax ; j++ ){
			trg_ptr[trg_index + j + from_col]
					= src_ptr[src_index + j];
		}
	}

	return 0;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_get_submatrix_by_index: ʬʥ٥ȥ˼Ф
 * ǽ
 *      ʥ٥ȥˤ饤ǥåǻꤵ줿ʬФ
 *
 *      
 *      1: ʥ٥ȥˤ饤ǥåǻꤵ줿ʬФ
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *trg;	O: ¤ΤؤΥݥ
 *      int nindex;		I: ǥå¤Το
 *      m_index_t *index;	I: ǥå¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_get_submatrix_by_index(m_matrix_t *src, m_matrix_t *trg, int nindex, m_index_t *index)
{
	int	by_i;
	int	*idx = NULL;
	double	*src_ptr, *trg_ptr;
	int	rowmax, colmax, i, j;

	by_i = 1;
	for( i = 0 ; i < nindex ; i++ ){
		if( index[i].j != 0 ){
			by_i = 0;
			break;
		}
	}
	if( !by_i ){
		for( i = 0 ; i < nindex ; i++ ){
			if( index[i].i != 0 ){
				goto err;
			}
		}
	}

	idx = (int *) malloc(nindex * sizeof(int));
	if( idx == NULL ){
		goto err;
	}

	for( i = 0 ; i < nindex ; i++ ){
		if( by_i ){
			idx[i] = index[i].i;
		}else{
			idx[i] = index[i].j;
		}
	}

	rowmax = src->row;
	colmax = src->col;
	src_ptr = src->elem;
	trg_ptr = trg->elem;
	if( rowmax == 1 || colmax == 1 ){
		/* ٥ȥ뤫Ф */
		for( i = 0 ; i < nindex ; i++ ){
			trg_ptr[i] = src_ptr[idx[i]];
		}
	}else{
		/* 󤫤Ф */
		for( i = 0 ; i < nindex ; i++ ){
			for( j = 0 ; j < nindex ; j++ ){
				trg_ptr[i*nindex + j] =
					src_ptr[idx[i] * colmax + idx[j]];
			}
		}
	}

	free(idx);

	return 0;
err:
	if( idx )	free(idx);
	log_sys("ERROR: mtx_get_submatrix_by_index()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_set_submatrix_by_index: ʬʥ٥ȥꡣ
 * ǽ
 *      ʥ٥ȥˤΥǥåǻꤵ줿ʬꤹ롣
 *
 *      
 *      1: ʥ٥ȥˤΥǥåǻꤵ줿ʬꤹ롣
 * 
 *      m_matrix_t *src;	I: ¤ΤؤΥݥ
 *      m_matrix_t *trg;	O: ¤ΤؤΥݥ
 *      int nindex;		I: ǥå¤Το
 *      m_index_t *index;	I: ǥå¤ΤؤΥݥ
 * ؿ
 *      0 :	
 *      -1:	۾
 * 
 *****************************************************************************/
int mtx_set_submatrix_by_index(m_matrix_t *src, m_matrix_t *trg, int nindex, m_index_t *index)
{
	int	by_i;
	int	*idx = NULL;
	double	*src_ptr, *trg_ptr;
	int	rowmax, colmax, i, j;

	by_i = 1;
	for( i = 0 ; i < nindex ; i++ ){
		if( index[i].j != 0 ){
			by_i = 0;
			break;
		}
	}
	if( !by_i ){
		for( i = 0 ; i < nindex ; i++ ){
			if( index[i].i != 0 ){
				goto err;
			}
		}
	}

	idx = (int *) malloc(nindex * sizeof(int));
	if( idx == NULL ){
		goto err;
	}

	for( i = 0 ; i < nindex ; i++ ){
		if( by_i ){
			idx[i] = index[i].i;
		}else{
			idx[i] = index[i].j;
		}
	}

	rowmax = src->row;
	colmax = src->col;
	src_ptr = src->elem;
	trg_ptr = trg->elem;
	if( rowmax == 1 || colmax == 1 ){
		/* ٥ȥꤹ롣 */
		for( i = 0 ; i < nindex ; i++ ){
			trg_ptr[idx[i]] = src_ptr[i];
		}
	}else{
		/* ꤹ롣 */
		for( i = 0 ; i < nindex ; i++ ){
			for( j = 0 ; j < nindex ; j++ ){
				trg_ptr[idx[i] * colmax + idx[j]] =
					src_ptr[i * nindex + j];
			}
		}
	}

	free(idx);

	return 0;
err:
	if( idx )	free(idx);
	log_sys("ERROR: mtx_set_submatrix_by_index()\n");
	return -1;
}

/*****************************************************************************
 * ؿ̾
 *      mtx_dump: ʥ٥ȥˤΥ
 * ǽ
 *      ʥ٥ȥǤפ롣
 *
 *      
 *      1: ʥ٥ȥǤפ롣
 * 
 *	int level;		I: ǥХå٥
 *      char *com;		I: ɽ
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *	̵
 * 
 *	ǥХå
 *****************************************************************************/
void mtx_dump(int level, char *com, m_matrix_t *mtx)
{
	mtx_dump_withname(level, com, NULL, mtx);
}
/*****************************************************************************
 * ؿ̾
 *      mtx_dump_withname: ʥ٥ȥˤΥ
 * ǽ
 *      ʥ٥ȥǤפ롣
 *
 *      
 *      1: ʥ٥ȥǤפ롣
 * 
 *	int level;		I: ǥХå٥
 *      char *com;		I: ɽ
 *      char *name;		I: ̾
 *      m_matrix_t *mtx;	I: ¤ΤؤΥݥ
 * ؿ
 *	̵
 * 
 *	ǥХå
 *****************************************************************************/
void mtx_dump_withname(int level, char *com, char *name, m_matrix_t *mtx)
{
	int	rowmax, colmax, i, j;
	double	*ptr;

	rowmax = mtx->row;
	colmax = mtx->col;
	ptr = mtx->elem;

	DBGW(level, ("%s %d x %d matrix\n", com, rowmax, colmax));
	for( i = 0 ; i < rowmax ; i++ ){
		if( name == NULL ){
			DBGW(level, ("[%2d] ", i));
		}else{
			DBGW(level, ("%s[%2d] ", name, i));
		}
		for( j = 0 ; j < colmax ; j++ ){
			DBGW(level, ("%.15e ", ptr[i*colmax+j]));
		}
		DBGW(level, ("\n"));
	}
}
