#ifndef _MATRIX_H_
#define _MATRIX_H_

#include	"ut_alloc.h"

typedef struct _matrix {
	int	row;		/* Ԥ	*/
	int	col;		/* 	*/
	double	*elem;		/* 	*/
} m_matrix_t;

typedef struct _index {
	int	i;		/* ԤΥǥå	*/
	int	j;		/* Υǥå	*/
} m_index_t;



#define MTX_SET(m, r, c, e)	{m.row = r; m.col = c; m.elem = e;}
#define MTX_SET_ALLOC(m, r, c)	{m.row = r; m.col = c; m.elem = (double *) malloc((r) * (c) * sizeof(double)); if(m.elem){memset(m.elem, 0, (r)*(c)*sizeof(double));}}
#define MTX_SET_FREE(m)	{if(m.elem){free(m.elem); m.row = m.col = 0; m.elem = NULL;} }

#define MTX_VEC	0
#define MTX_COL	1
#define MTX_ROW	2

#define PLUS_INF	0
#define MINUS_INF	-1

#define MTX_EQ		0
#define MTX_NOTEQ	1
#define MTX_LESS	2
#define MTX_LESSEQ	3
#define MTX_MORE	4
#define MTX_MOREEQ	5



extern int mtx_isdiagonal(m_matrix_t *mtx);
extern int mtx_issparse(m_matrix_t *mtx);
extern int mtx_unit(m_matrix_t *mtx);
extern int mtx_unique(double s, m_matrix_t *mtx);
extern int mtx_nonzero(m_matrix_t *mtx);
extern int mtx_elem_count(double val, m_matrix_t *mtx);
extern int mtx_any(m_matrix_t *mtx, int relation, double val);
extern int mtx_index(m_matrix_t *mtx, int relation, double val, m_index_t **indexpp);
extern void mtx_index_free(m_index_t *index);
extern int mtx_set_elem(double val, int len, m_index_t *index, m_matrix_t *mtx);
#if 0
extern double mtx_est_condition_number(m_matrix_t *mtx);
#endif

extern int mtx_transpose(m_matrix_t *src, m_matrix_t *trg);
extern int mtx_addup(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX);
extern int mtx_subtract(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX);
extern int mtx_multiply(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX);

extern int mtx_elem_multiply(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX);
extern int mtx_elem_divide(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX);
extern int mtx_elem_power(m_matrix_t *MA, m_matrix_t *MB, m_matrix_t *MX);

extern int mtx_elem_addup_to_scalar(double p, m_matrix_t *MA, m_matrix_t *MX);
extern int mtx_elem_subtract_from_scalar(double p, m_matrix_t *MA, m_matrix_t *MX);
extern int mtx_elem_multiply_by_scalar(double p, m_matrix_t *MA, m_matrix_t *MX);
extern int mtx_elem_divide_into_scalar(double p, m_matrix_t *MA, m_matrix_t *MX);

extern int mtx_scalar_addup(m_matrix_t *MA, double p, m_matrix_t *MX);
extern int mtx_scalar_multiply(m_matrix_t *MA, double p, m_matrix_t *MX);
extern int mtx_scalar_power(m_matrix_t *MA, double p, m_matrix_t *MX);

extern int mtx_diagonal(m_matrix_t *vec, m_matrix_t *diag);
extern int mtx_diagonal2vector(int k, m_matrix_t *diag, m_matrix_t *vec);

extern int mtx_sum(int sumdir, m_matrix_t *mtx, m_matrix_t *sum);
extern int mtx_abs(m_matrix_t *src, m_matrix_t *abs);
extern int mtx_min(m_matrix_t *src, m_matrix_t *min);
extern int mtx_max(m_matrix_t *src, m_matrix_t *max);
extern int mtx_max_index(m_matrix_t *src, m_matrix_t *max, int index[]);
extern int mtx_max_row(m_matrix_t *src, m_matrix_t *max);
extern int mtx_max2(m_matrix_t *src0, m_matrix_t *src1, m_matrix_t *max);

extern int mtx_cumprod(m_matrix_t *mtx, int dim, m_matrix_t *cum);
extern int mtx_norm(m_matrix_t *mtx, int p, double *norm);

extern int mtx_get_submatrix(m_matrix_t *src, m_matrix_t *trg, int from_row, int from_col);
extern int mtx_set_submatrix(m_matrix_t *src, m_matrix_t *trg, int from_row, int from_col);
extern int mtx_get_submatrix_by_index(m_matrix_t *src, m_matrix_t *trg, int nindex, m_index_t *index);
extern int mtx_set_submatrix_by_index(m_matrix_t *src, m_matrix_t *trg, int nindex, m_index_t *index);

extern void mtx_dump(int level, char *com, m_matrix_t *mtx);
extern void mtx_dump_withname(int level, char *com, char *name, m_matrix_t *mtx);

#endif /* _MATRIX_H_ */
