#include	<stdio.h>
#include	<math.h>

#include	"logUtil.h"

#include	"nr_lu.h"
#include	"nr_mutil.h"

#define	TINY	1.0e-20


/*----------------nr_ludcmp------------------------------------------
 *
 *  LUʬ
 *
 *      I:a, n, indx
 *      O:a, indx, d
 */
int nr_ludcmp(double **a, int n, int *indx, double *d)
{
        int i,j,k,imax=0;
        double big,dum,sum,temp;
        double *vv;

	vv = nr_alloc_vector(1, n);
	if( vv == NULL ){
		log_sys("nr_alloc_vector() ERROR\n");
		goto err;
	}

        *d=1.0;
        for (i=1;i<=n;i++) {
                big=0.0;
                for (j=1;j<=n;j++)
                        if ((temp=fabs(a[i][j])) > big) big=temp;
                if (big == 0.0) {
                        log_sys("## Singular matrix in nr_ludcmp\n");
                        goto err_singular;
                }
                vv[i]=1.0/big;

        }
        for (j=1;j<=n;j++) {
                for (i=1;i<j;i++) {
                        sum=a[i][j];
                        for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
                        a[i][j]=sum;
                }
                big=0.0;
                for (i=j;i<=n;i++) {
                        sum=a[i][j];
                        for (k=1;k<j;k++)
                                sum -= a[i][k]*a[k][j];
                        a[i][j]=sum;
                        if ( (dum=vv[i]*fabs(sum)) >= big) {
                                big=dum;
                                imax=i;
                        }
                }
#ifdef DBG
                if(imax==0) {
                        /* DBGW(1, ("##Error in ludcmp imax=%d\n")); */
                        goto err;
                }
#endif
                if (j != imax) {
                        for (k=1;k<=n;k++) {
                                dum=a[imax][k];
                                a[imax][k]=a[j][k];
                                a[j][k]=dum;
                        }
                        *d = -(*d);
                        vv[imax]=vv[j];
                }
                indx[j]=imax;
                if (a[j][j] == 0.0) a[j][j]=TINY;
                if (j != n) {
                        dum=1.0/(a[j][j]);
                        for (i=j+1;i<=n;i++) a[i][j] *= dum;
                }
        }

	nr_free_vector(vv, 1, n);
	return 0;
err:
	if( vv )	nr_free_vector(vv, 1, n);
	log_sys("ERROR: nr_ludcmp()\n");
	return -1;
err_singular:
	if( vv )	nr_free_vector(vv, 1, n);
	return -2;
}

/*----------------nr_lubksb------------------------------------------
 *
 *  LU
 *  n Ĥ A*x = b 򤯡
 *  Ϥ a[1..n][1..n] ϡludcmp() ֤ A LUʬǤ롣
 *  Ϥ٥ȥ b[1..n] ϡդΥ٥ȥǤ롣
 *  ܥ롼ϡa, n, indx 񤭴ʤΤǡ
 *  դ b ѤäƤ֤ⷫȤ롣
 *  b κǽˡ¤ǤǽθƤΤǡ
 *  ΨŪ˵չ󤬵롣
 *
 *      I:a, n, indx, b
 *      O:b
 *      ꥿͡ʤ
 */
void nr_lubksb(double **a, int n, int *indx, double b[])
{
        int	i, ii = 0, ip ,j;
        double	sum;


	/*
	 * ii ʤ顢ϡb κǽΥǤʤǤźǤ롣
	 * Ը򴹤򸵤ᤷʤ鼰Ԥ
	 * A*x = (L*U)*x = L*(U*x) = b
	 * L*y = b, U*x = y
	 * y[1] = b[1] / L[1,1]
	 * y[i] = {b[i] - Sigma(j=1,i-1)(L[i,j]*y[j])} / L[i,i]
	 * i = 2, 3, ..., n; L[i,i] = 1
	 */
        for ( i = 1 ; i <= n ; i++ ) {
                ip = indx[i];
                sum = b[ip];
                b[ip] = b[i];
                if (ii){
                        for ( j = ii ; j <= i - 1 ; j++ ){
				sum -= a[i][j] * b[j];
			}
		}else if (sum){
			/*
			 * 󥼥Ǥ˽вä
			 * ʹߤϾΥ롼פ¤롣
			 */
			ii = i;
		}
                b[i] = sum;
        }
	/*
	 * θԤ
	 * x[n] = y[n] / U[n,n]
	 * x[i] = {y[i] - Sigma(j=i+1,n)(U[i,j] * x[j])} / U[i,i]
	 * i = n-1, n-2, ..., 1
	 */
        for ( i = n ; i >= 1 ; i-- ) {
                sum = b[i];
                for ( j = i + 1 ; j <= n ; j++ ){
			sum -= a[i][j] * b[j];
		}
                b[i] = sum / a[i][i];
        }
}


/*----------------nr_tlubksb------------------------------------------
 *
 *  LUžǡ
 *  n Ĥ A'*x = b 򤯡A'ϡAžֹ
 *  Ϥ a[1..n][1..n] ϡludcmp() ֤ A LUʬǤ롣
 *  Ϥ٥ȥ b[1..n] ϡդΥ٥ȥǤ롣
 *  ܥ롼ϡa, n, indx 񤭴ʤΤǡ
 *  դ b ѤäƤ֤ⷫȤ롣
 *  b κǽˡ¤ǤǽθƤΤǡ
 *  ΨŪ˵չ󤬵롣
 *
 *      I:a, n, indx, b
 *      O:b
 *      ꥿͡ʤ
 */
int nr_tlubksb(double **a, int n, int *indx, double b[])
{
        int	i, ii = 0, ip ,j;
        double	sum;
	double	*y	= NULL;

	y = nr_alloc_vector(1, n);
	if( y == NULL){
		log_sys("nr_alloc_vector() ERROR\n");
		goto err;
	}

	/*
	 * ii ʤ顢ϡb κǽΥǤʤǤźǤ롣
	 * Ԥ
	 * A'*x = (L*U)'*x = (U'*L')*x = U'*(L'*x) = b
	 * U'*y = b, L'*x = y
	 * y[1] = b[1] / U'[1,1]
	 * y[i] = {b[i] - Sigma(j=1,i-1)(U'[i,j]*y[j])} / U'[i,i]
	 * i = 2, 3, ..., n
	 */
        for ( i = 1 ; i <= n ; i++ ) {
                sum = b[i];
                if (ii){
                        for ( j = ii ; j <= i - 1 ; j++ ){
				sum -= a[j][i] * y[j];
			}
		}else if (sum){
			/*
			 * 󥼥Ǥ˽вä
			 * ʹߤϾΥ롼פ¤롣
			 */
			ii = i;
		}
                y[i] = sum / a[i][i];
        }
	/*
	 * Ը򴹤򸵤ᤷʤ鼰θԤ
	 * x[n] = y[n] / L'[n,n]
	 * x[i] = {y[i] - Sigma(j=i+1,n)(L'[i,j] * x[j])} / L'[i,i]
	 * i = n-1, n-2, ..., 1; L'[i,i] = 1
	 */
        for ( i = n ; i >= 1 ; i-- ) {
                sum = y[i];
                for ( j = i + 1 ; j <= n ; j++ ){
			sum -= a[j][i] * y[j];
		}
		y[i] = sum;
                ip = indx[i];
		b[i] = b[ip];
                b[ip] = sum;
        }


	nr_free_vector(y, 1, n);
	return 0;
err:
	if( y )	nr_free_vector(y, 1, n);
	log_sys("ERROR: nr_tlubksb()\n");
	return -1;
}

