#define _HAVE_MAIN_ROUTINE_
#include "life.h"
#include "rs_io.h"
#include "rs_utl.h"
#include "utl_rand.h"
#include "utl_alloc.h"
#include "lsUsrFitness.h"

#ifndef NFS
#include "utl_mpi_ft.h"
#endif /* NFS */

#ifdef MPI
	#include <libgen.h>
	#include "mpi.h"
#endif

#ifdef WIN32
#include "getopt.h"
#else
#include <unistd.h>
#endif

/***** STATIC VARIABLES *****/
static char *fName[FILE_NUM];		/* ϥե̾ */
static FILE *fpOutList;				/* ̰νϥեݥ */
static int fListSort = F_LIST_SORT;	/* 򥽡Ȥ뤫ݤΥե饰 */

/***** STATIC FUNCTIONS *****/

/*  */
static int my_init(int argc, char *argv[], int myid, int numprocs, int *num_search, int *job_unit);

/* λ */
static void my_final(int myid);

/* ñץǤõԤ */
static int do_search_single(Parameter *default_param, double *y_default, double h, double t_start, double t_end, double dt_save, int num_search);

#ifdef MPI
/* MPIѤʣץǤõԤ */
static int do_search_mpi(Parameter *default_param, double *y_default, double h, double t_start, double t_end, double dt_save, int num_search, int job_unit, int mpi_rank, int numprocs);
#endif /* MPI */


/***** MAIN *****/
int main(int argc, char *argv[])
{
	int i;
    int myid=0, numprocs=1;
	int	num_search, job_unit;
	int	rv = -1;
	double t_start, t_end, h, dt_save;
	double y_default[MAX_VAR_NUM];
	Parameter default_param;
#ifdef MPI
    int namelen;
    char processor_name[MAX_LINE];

    MPI_Init(&argc,&argv);
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    MPI_Get_processor_name(processor_name,&namelen);
#endif

	/* եΥץϥǡɤ߹ */
	if (my_init(argc, argv, myid, numprocs, &num_search, &job_unit) < 0) {
		goto out;
	}
	if (Set_data(y_default, &t_start, &t_end, &h, &dt_save, fName) < 0) {
		goto out;
	}
	if (AllocSolutionMatrix(t_start, t_end, dt_save) < 0) {
		goto out;
	}
	if (InitOutfileList(num_search, P_GaNum, myid) < 0) {
		goto out;
	}	
	
	for(i = 0; i < P_GaNum; i++){
		if(!strcmp("constantPlayer", Pno2Name(P_Ga[i].p_no))){
			log_sys(" ## Error: constantPlayer cannot be set as subject to search.\n");
			goto out;
		}
	}
	default_param = Param;
	
#ifndef DBG
#ifdef MPI
	if(DistributeSlaveSeed(myid, numprocs) < 0){
		goto out;
	}
	if(myid != 0) unlink(fName[FILE_IN]);	
#else /* MPI */
	SetSeedByTime();
#endif /* MPI */
#endif /* DBG */
	
	/* ϥե˷׻ѥ᡼ͤ񤭽Ф */
	if(myid != 0 || numprocs == 1) 
			WriteInputData(y_default, t_start, t_end, h, dt_save);

#ifdef MPI
    fprintf(stdout,"##### CPU %d/%d (%s)\n", myid+1, numprocs, processor_name);

	if(numprocs == 1){
		if(do_search_single(&default_param, y_default, h, t_start, t_end, 
								dt_save, num_search) == -1){
			goto out;
		}
	} else {
		if(do_search_mpi(&default_param, y_default, h, t_start, t_end, 
					dt_save, num_search, job_unit, myid, numprocs) == -1){
			goto out;
		}
	}
#else /* MPI */
	if(do_search_single(&default_param, y_default, h, t_start, t_end, 
							dt_save, num_search) == -1){
		goto out;
	}
#endif /* MPI */
	
	rv = 0;
out:
	if(myid == 0){
		if(rv == 0){
			printf("Finish Successfully.\n");
		} else {
			printf("Exit Error!!\n");
		}
	}
	my_final(myid);

#ifdef MPI
	if(rv == -1){
		MPI_Abort(MPI_COMM_WORLD, -1);
	}
	
	MPI_Finalize();
#endif
	return rv;
}


/*---------------- my_init------------------------------------------
 *
 * եΥץϥǡɤ߹νԤ
 *
 * I:argc, argv
 * O:y, t_start, t_end, h, dt_save
 *
 */
static int
my_init(int argc, char *argv[], int myid, int numprocs, int *num_search, int *job_unit)
{
#define FLAG_I	(1 << 0)
#define FLAG_O	(1 << 1)
#define FLAG_n	(1 << 2)
#define FLAG_j	(1 << 3)
#define FLAG_r	(1 << 4)
	int ch;
	int	rv = -1;
	char input_file[FILENAME_MAX];
	char myid_file[FILENAME_MAX];
	char list_file[FILENAME_MAX];
	unsigned int	foptions = 0;
	
#ifdef DBG
	// open debug-log file.
	if(log_dbginit(argc, argv)) {
		printf("Cannot open dbgLog file!!");
		goto	out;
	}
	DBGW(10, ("my_init enter.\n"));
#endif

	if(log_sysinit(argc, argv)) {
		printf("Cannot open sysLog file!!");
		goto	out;
	}

#ifdef DBG
	DBGW(10, ("my_init enter\n"));
#endif

	/* ץβ
	 * -I [LSϥե̾] (ɬ)			
	 * -O [ϥե̾] (ɬ)
	 * -n [õ] (ɬ)
	 * -j [졼֤˰ꤹ른֤ñ] (MPIΤߡꤷʤ1)
	 * -r ե򥽡Ȥʤɽ
	 */
	opterr = 0;
	while((ch = getopt(argc, argv, "I:O:n:j:r")) != EOF){
		switch(ch){
		case 'I':
			if(foptions & FLAG_I) {
				log_sys(" ## Error: Duplicated option: \'%c\'\n", ch);
				goto out;
			}
			strcpy(input_file, optarg);
			foptions |= FLAG_I;
			break;
		case 'O':
			if(foptions & FLAG_O) {
				log_sys(" ## Error: Duplicated option: \'%c\'\n", ch);
				goto out;
			}
			fName[FILE_OUT] = strdup(optarg);
			foptions |= FLAG_O;
			break;
		case 'n':
			if(foptions & FLAG_n) {
				log_sys(" ## Error: Duplicated option: \'%c\'\n", ch);
				goto out;
			}
			*num_search = atoi(optarg);
			foptions |= FLAG_n;
			break;
		case 'j':
			if(foptions & FLAG_j) {
				log_sys(" ## Error: Duplicated option: \'%c\'\n", ch);
				goto out;
			}
			*job_unit = atoi(optarg);
			foptions |= FLAG_j;
			break;
		case 'r':
			if(foptions & FLAG_r) {
				log_sys(" ## Error: Duplicated option: \'%c\'\n", ch);
				goto out;
			}
			fListSort = F_LIST_UNSORT;
			foptions |= FLAG_r;
			break;
		default:
			break;
		}
	}

	/* ɬܥץΥ顼å */	
	if(!(foptions & FLAG_I)){
		log_sys("my_init: input file not specified.\n");
		goto out;
	}
	if(!(foptions & FLAG_O)){
		log_sys("my_init: output file not specified.\n");
		goto out;
	}
	if(!(foptions & FLAG_n)){
		log_sys("my_init: number of search not specified.\n");
		goto out;
	}

	/* -jץ󤬻ꤵʤäΥǥե */
	if(!(foptions & FLAG_j)){
		*job_unit = 1;
	}

#ifdef MPI
#ifdef NFS
	fName[FILE_IN] = strdup(input_file);
#else /* NFS */

	/* ϥե(ޥ->졼) */
	if(myid == 0){
		fName[FILE_IN] = strdup(input_file);
		if(BcastFileSend(input_file) == -1){
			goto out;
		}
	} else {
		char in_file[FILENAME_MAX];
		sprintf(in_file, "%s_%d", basename(input_file), myid);
		fName[FILE_IN] = strdup(in_file);
		if(BcastFileRecv(in_file) == -1){
			goto out;
		}
	}

	// ϥե򥪡ץ󤹤
	if(myid != 0){
		sprintf(myid_file, "%s_tmp_%d", basename(fName[FILE_OUT]), myid);
		if((Fp_out = fopen(myid_file, "w")) == NULL){
			log_sys("my_init: outputfile[%s] open err!!\n", myid_file);
			goto out;
		}
	}
#endif /* NFS */
#else /* MPI */
	fName[FILE_IN] = strdup(input_file);
#endif /* MPI */
	
	// ϥե򥪡ץ󤹤
	if(myid == 0){
		if(numprocs == 1){
			if((Fp_out = fopen(fName[FILE_OUT], "w")) == NULL){
				log_sys("my_init: outputfile[%s] open err!!\n", fName[FILE_OUT]);
				goto out;
			}
		}
		sprintf(list_file, "%s_list", fName[FILE_OUT]);
		if((fpOutList = fopen(list_file, "w")) == NULL){
			log_sys("my_init: outputfile[%s] open err!!\n", list_file);
			goto out;
		}
	}

	rv = 0;
out:
#ifdef DBG
	DBGW(10, ("my_init exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

static void my_final(int myid)
{
#ifdef DBG
	DBGW(10, ("my_final enter.\n"));
#endif /* DBG */

	if(SolutionMatrix != NULL) FreeDoubleMatrix(SolutionMatrix);
	
	/*
	 * ѥ᡼γǼΰγ
	 */
	freeParamTag();

	/*
	 * ϥեĤ
	 */
	if( Fp_out ){
		fclose(Fp_out); Fp_out = NULL;
	}
	if( fpOutList != NULL){
		fclose(fpOutList);
	}

#ifdef MPI
	if(myid != 0){
		unlink(fName[FILE_IN]);
	}
#endif /* MPI */
	
	/*
	 * λ
	 */
	log_sysfinalize();
#ifdef DBG
	DBGW(10, ("my_final exit\n"));
#endif /* DBG */

	log_dbgfinalize();
}


static int do_search_single(Parameter *default_param, double *y_default, double h, double t_start, double t_end, double dt_save, int num_search)
{
	int	i, i_search;
	int	ret;
	int	rv = -1;
	double	fitness;
	double 	y[MAX_VAR_NUM], y_start[MAX_VAR_NUM], y_pre[MAX_VAR_NUM];
	double	random_vector[MAX_P_GA];
	Parameter param_start;

#ifdef DBG
	DBGW(10, ("do_search_single enter.\n"));
#endif /* DBG */
	
	/* ѥ᡼٥Υ롼 */
	for (i_search = 0; i_search < num_search; i_search++) {
		ParamB = param_start = Param = *default_param;
		for (i=1;i<=Param.n_var;i++) y_start[i] = y_pre[i] = y[i] = y_default[i];
		SetRandomParam(random_vector);
		
		StepNum = 0;
		ret = DoSolver(param_start, h, t_start, t_end, dt_save, y_start, y, y_pre);
		if(LsUsrFitness(SolutionMatrix, StepNum, Param.n_var, ret, &fitness) == -1){
			goto out;
		}
		PrintResultFile(random_vector, fitness, 0, i_search, ret);
	}

	
	PrintList(fpOutList, num_search, fListSort);
	
	rv = 0;
out:
#ifdef DBG
	DBGW(10, ("do_search_single exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

#ifdef MPI

#define CMD_MPI_END -1			/* 졼֤ؤνλ̿ */
#define CMD_MPI_FILE_GATHER -2	/* 졼֤ؤΥեž̿ */

/***** STATIC FUNCTIONS for do_search_mpi *****/
static int master_job(int num_search, int job_unit, int numprocs);
static int slave_job(Parameter *default_param, double *y_default, double h, double t_start, double t_end, double dt_save, int myid);

static int do_search_mpi(Parameter *default_param, double *y_default, double h, double t_start, double t_end, double dt_save, int num_search, int job_unit, int mpi_rank, int numprocs)
{
	int rv = -1;

#ifdef DBG
	DBGW(10, ("do_search_mpi enter.\n"));
#endif /* DBG */

	if(mpi_rank == 0){
		if(master_job(num_search, job_unit, numprocs) < 0){
			goto out;
		}
	} else {
		if(slave_job(default_param, y_default, h, t_start, t_end, dt_save, mpi_rank) < 0){
			goto out;
		}
	}

	rv = 0;
out:
#ifdef DBG
	DBGW(10, ("do_search_mpi exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}


static int master_job(int num_search, int job_unit, int numprocs)
{
	int	i;
	int	rv = -1;
	int head = 0;
	int num_send = 0;
	int num_recv = 0;
	int	num_search_last;		/* ֥塼ǸΥñ */
	int *job_queue;
	int	queue_size;
	int	send_message;
	MPI_Status stat;

#ifdef DBG
	DBGW(10, ("master_job enter.\n"));
#endif /* DBG */
	
	/* ֥塼ν */
	queue_size = num_search/job_unit;	
	num_search_last = num_search%job_unit;
	if(num_search_last != 0){
		queue_size++;
	} else {
		num_search_last = job_unit;
	}
	if((job_queue = IntVectorAlloc(queue_size)) == NULL){
		goto out;
	}
	for(i = 0; i < queue_size - 1;i++){
		job_queue[i] = job_unit;
	}
	job_queue[i] = num_search_last;	
	
	/* ֤μ¹԰ */
	for(i = 0; i < numprocs-1; i++){
		if(num_send == queue_size) break;
		if(MPI_Send(&(job_queue[num_send]), 1, MPI_INT, i + 1, 0, 
							   MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In master_job1).\n");
			goto out;
		}
		num_send++;	
	}
	for(;;){
		int recv_message;
		if(MPI_Recv(&recv_message, 1, MPI_INT, MPI_ANY_SOURCE,
					MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Recv(In master_job).\n");
			goto out;
		}
		num_recv++;
		if(num_send < queue_size){
			int next_dest;
			next_dest = stat.MPI_SOURCE;
			if(MPI_Send(&(job_queue[num_send]), 1, MPI_INT, next_dest, 0, 
							   MPI_COMM_WORLD) != MPI_SUCCESS){
				log_sys(" ## Error: MPI_Send(In master_job2).\n");
				goto out;
			}
			num_send++;
		}
		if(num_recv == queue_size) break;
	}

	/* ե롦μ */
	send_message = CMD_MPI_FILE_GATHER;
	for(i = 0; i < numprocs - 1; i++){
		int recv_size;
		char file_name[FILENAME_MAX];
		if(MPI_Send(&send_message, 1, MPI_INT, i + 1, 0, 
								MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In master_job3).\n");
			goto out;
		}
		sprintf(file_name, "%s_%d", fName[FILE_OUT], i+1);
		RecvFile(file_name);
		recv_size = RecvList(head, P_GaNum);
		head += recv_size;
	}

	/* 졼֤˽λ̿ */	
	send_message = CMD_MPI_END;
	for(i = 0; i < numprocs - 1; i++){
		if(MPI_Send(&send_message, 1, MPI_INT, i + 1, 0, 
								MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In master_job3).\n");
			goto out;
		}
	}

	/* եؤν */
	PrintList(fpOutList, num_search, fListSort);
	
	rv = 0;
out:
	if(job_queue != NULL) FreeIntVector(job_queue);

#ifdef DBG
	DBGW(10, ("master_job exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
		
static int slave_job(Parameter *default_param, double *y_default, double h, double t_start, double t_end, double dt_save, int myid)
{
	int i, j;
	int	ret, rv = -1;
	int total_search = 0;
	int num_search_once;
	int message;
	double fitness;
	double y[MAX_VAR_NUM], y_pre[MAX_VAR_NUM], y_start[MAX_VAR_NUM];
	double random_vector[MAX_P_GA];
	Parameter param_start;
	MPI_Status stat;

#ifdef DBG
	DBGW(10, ("slave_job exit(%d)\n", rv));
#endif /* DBG */
	
	for(;;){
		/* ޥ̿ */
		if(MPI_Recv(&num_search_once, 1, MPI_INT, 0, MPI_ANY_TAG, 
								MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Recv(In slave_job).\n");
			goto out;
		}

		/* λ̿ */
		if(num_search_once == CMD_MPI_END) break;

		/* ̤ޥ */
		if(num_search_once == CMD_MPI_FILE_GATHER){
			char file_name[FILENAME_MAX];
			sprintf(file_name, "%s_tmp_%d", fName[FILE_OUT], myid);
			fclose(Fp_out);
			
			/* ե */
			if(SendFile(file_name, 0) < 0) goto out;
			
			/*  */
			if(SendList(total_search, P_GaNum, 0) < 0) goto out; 

			/* ե */
			unlink(file_name);
			
			continue;
		}

		/* õ¹Ԥλåޥ */
		for(i = 0; i < num_search_once; i++){
			ParamB = param_start = Param = *default_param;
			for(j = 0; j < Param.n_var; j++) 
					y_start[i] = y_pre[i] = y[i] = y_default[i];
			
			SetRandomParam(random_vector);

			StepNum = 0;
			ret = DoSolver(param_start, h, t_start, t_end, dt_save, 
							y_start, y, y_pre);
			if(LsUsrFitness(SolutionMatrix, StepNum, Param.n_var + 1, 
									ret, &fitness) < 0){
				goto out;
			}
			PrintResultFile(random_vector, fitness, myid, total_search, ret);
			total_search++;
		}
		message = 0;
		if(MPI_Send(&message, 1, MPI_INT, 0, 0, MPI_COMM_WORLD) != MPI_SUCCESS){
			log_sys(" ## Error: MPI_Send(In slave_job).\n");
			goto out;
		}
	}
	
	rv = 0;
out:
#ifdef DBG
	DBGW(10, ("slave_job exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}	
#endif /* MPI */
#undef _HAVE_MAIN_ROUTINE_
