/* *****************************************************************************
 * 
 *	S-system
 *
 */

#include "life.h"
#include "nr_lu.h"
#include "nr_mutil.h"

/* եstaticؿΥץȥ */
static int allocate_base_matrixes(double ***Gd, double ***Gi, double ***Hd, double ***Hi,
		    	int n_var, int n_const);
static void free_base_matrixes(double ***Gd, double ***Gi, double ***Hd, double ***Hi,
			   	int n_var, int n_const);
static int std_analysis(FILE *Fp_out, double **Gd, double **Gi, double **Hd, double **Hi,
			   	double *y, int n_var, int n_const);
static void sens_kinetic(FILE *Fp_out, double **GHd, double **GHi, double **Ad_inv,
			   	double *y, int n_var, int n_const, int f_isG);
static void sens_kinetic_flux(FILE *Fp_out, double **GHd, double **GHi, double **Ad_inv,
			   	double *y, int n_var, int n_const, int f_isG);

static void set_params(double **Gd, double **Gi, double **Hd, double **Hi,
				int n_var, int n_const);
static int get_param(int p_no, int p_idx, int *idx1, int *idx2);

/*------------------ SsysSTD --------------------------------------
 *	S-systemˤ롣
 */
int SsysSTD(double *y)
{
	int i, j;
	int n_var, n_const;
	double **Gd=NULL, **Gi=NULL, **Hd=NULL, **Hi=NULL;

	n_var = Param.n_var;
	n_const = Param.n_val[1];

	/* ܤȤʤޥȥå٥ȥΰ */
	if (allocate_base_matrixes(&Gd, &Gi, &Hd, &Hi, n_var, n_const)) goto err;

	for(i=1; i<=n_var; i++ ){
		for( j=1; j<=n_var; j++ ) Gd[i][j] = Hd[i][j] = 0.0;
		for( j=1; j<=n_const; j++ ) Gi[i][j] = Hi[i][j] = 0.0;
	}

	/*
	 *	ѥ᡼Υ򸵤G,H򥻥åȤ롣
	 */
	set_params(Gd, Gi, Hd, Hi, n_var, n_const);

	/*
	 *	ٲϤԤη̤Ϥ롣
	 */
	if (std_analysis(Fp_out, Gd, Gi, Hd, Hi, y, n_var, n_const)) goto err;

	/* ܤȤʤޥȥå٥ȥΰ賫 */
	free_base_matrixes(&Gd, &Gi, &Hd, &Hi, n_var, n_const);
	
	return 0;

err:
	log_sys("\n## Error in SsysSTD.\n");

	/* ܤȤʤޥȥå٥ȥΰ賫 */
	free_base_matrixes(&Gd, &Gi, &Hd, &Hi, n_var, n_const);

	return -1;
}

/*---------------------- set_params -------------------------------------
 *	ޥȥå˥ѥ᡼򥻥åȤ롣
 */
static void
set_params(double **Gd, double **Gi, double **Hd, double **Hi,
				int n_var, int n_const)
{
	int i, idx_plus, idx_minus, idx1, idx2;

	idx_plus = Name2Pno("g");
	idx_minus = Name2Pno("h");
	for(i=1; i<=Param.n_val[idx_plus]; i++ ) {
		if (get_param(idx_plus, i, &idx1, &idx2)) {
			Gd[idx1][idx2] = Param.val[idx_plus][i];
		} else {
			Gi[idx1][idx2] = Param.val[idx_plus][i];
		}
	}
	for(i=1; i<=Param.n_val[idx_minus]; i++ ) {
		if (get_param(idx_minus, i, &idx1, &idx2)) {
			Hd[idx1][idx2] = Param.val[idx_minus][i];
		} else {
			Hi[idx1][idx2] = Param.val[idx_minus][i];
		}
	}
}

/*---------------------- get_param -------------------------------------
 *	ѥ᡼ΥˤΥǥå롣
 *
 *	֤ dependent:1 independent:0
 */
static int
get_param(int p_no, int p_idx, int *idx1, int *idx2)
{
	char *str_tmp, tline[MAX_LINE];
	int f_var;

	str_tmp = getParamTag(p_no, p_idx);
	strcpy(tline, str_tmp);
	str_tmp = tline;
	if (strstr(str_tmp, "d")) {
		f_var = 1;
	} else {
		f_var = 0;
	}

	strsep(&str_tmp, "_");
	*idx1 = atoi(strsep(&str_tmp, "_"));
	*idx2 = atoi(strsep(&str_tmp, "_"));

	return f_var;
}

/*---------------------- std_analysis -------------------------------------
 *	֤ˤƼϤȤν
 */
static int
std_analysis(FILE *Fp_out, double **Gd, double **Gi, double **Hd, double **Hi,
			   	double *y, int n_var, int n_const)
{
	int i, j, k;
	int *index=NULL;
	double **Ad=NULL, **Ad_inv=NULL, **Ai=NULL;
	double **Ad1Ai=NULL, **GdAd1Ai=NULL;
	double *col=NULL, *alpha, *beta, *const_var, *beta_save=NULL, *flux=NULL;
	double d, val;

	Ad = nr_alloc_matrix(1, n_var, 1, n_var);
	if (Ad == NULL) goto err_alloc;
	Ad_inv = nr_alloc_matrix(1, n_var, 1, n_var);
	if (Ad_inv == NULL) goto err_alloc;
	Ai = nr_alloc_matrix(1, n_var, 1, n_const);
	if (Ai == NULL) goto err_alloc;
	Ad1Ai = nr_alloc_matrix(1, n_var, 1, n_const);
	if (Ad1Ai == NULL) goto err_alloc;
	GdAd1Ai = nr_alloc_matrix(1, n_var, 1, n_const);
	if (GdAd1Ai == NULL) goto err_alloc;
	col = nr_alloc_vector(1, n_var);
	if (col == NULL) goto err_alloc;
	beta_save = nr_alloc_vector(1, n_var);
	if (beta_save == NULL) goto err_alloc;
	flux = nr_alloc_vector(1, n_var);
	if (flux == NULL) goto err_alloc;
	index = nr_alloc_ivector(1, n_var);
	if (index == NULL) goto err_alloc;

	alpha = Param.val[Name2Pno("kplus")];
	beta = Param.val[Name2Pno("kminus")];
	const_var = Param.val[Name2Pno(CONSTANT_PLAYER)];

	nr_sub_matrix(Gd, Hd, Ad, n_var, n_var);
	nr_sub_matrix(Gi, Hi, Ai, n_var, n_const);

	if (nr_ludcmp(Ad, n_var, index, &d) < 0) goto err;
	
	/* չAd_inv η׻ */
	for (i=1; i<=n_var; i++) {
		for (j=1; j<=n_var; j++) col[j] = 0.0;
		col[i] = 1.0;
		nr_lubksb(Ad, n_var, index, col);
		for (j=1; j<=n_var; j++) Ad_inv[j][i] = col[j];
	}

	nr_mult_matrix(Ad_inv, Ai, Ad1Ai, n_var, n_var, n_const);

	/* η׻ */
	for (i=1; i<=n_var; i++) {
		val = 0.0;
		for (j=1; j<=n_var; j++) val += Ad_inv[i][j]*log(beta[j]/alpha[j]);
		for (j=1; j<=n_const; j++) val -= Ad1Ai[i][j]*log(const_var[j]);
		y[i] = exp(val);
	}

	/* 
	 * flux η׻
	 * beta Ū˥ˤfvec׻Ƶ롣
	 */
	for(i=1; i<=n_var; i++) {
		beta_save[i] = beta[i];
		beta[i] = 0.0;
	}
	usr_fvec(flux, y, GeneConc, &Param);
	for(i=1; i<=n_var; i++) beta[i] = beta_save[i];
			
	fprintf(Fp_out, "\n$ STD Values and Fluxes $\n");
	fprintf(Fp_out, "index,  STD value ,      Flux\n");
	for(i=1; i<=n_var; i++) {
		fprintf(Fp_out, "%5d,%12.3e,%12.3e, #%s\n",
				i, y[i]/GeneConc, flux[i], getVarTag(i));
	}
	
	nr_mult_matrix(Gd, Ad1Ai, GdAd1Ai, n_var, n_var, n_const);

	/* ѿФ봶 */
	fprintf(Fp_out,"\n$ Logarythmic Gains of Metabolites $\n");
	fprintf(Fp_out,"%-19s", "param/var");
	for (i=1; i<=n_var; i++) fprintf(Fp_out,",     y[%3d]", i); 
	fprintf(Fp_out,"\n");
	for (j=1; j<=n_const; j++) {
		fprintf(Fp_out,"%14s[%3d]", CONSTANT_PLAYER, j);
		for (i=1; i<=n_var; i++) fprintf(Fp_out, ",%11.3e", -Ad1Ai[i][j]); 
		fprintf(Fp_out,"\n");
	}
	
	fprintf(Fp_out,"\n$ Sensitivities of Metabolites with respect to rate constants $\n");
	fprintf(Fp_out,"%-19s", "param/var");
	for (i=1; i<=n_var; i++) fprintf(Fp_out,",     y[%3d]", i); 
	fprintf(Fp_out,"\n");
	for (j=1; j<=n_var; j++) {
		fprintf(Fp_out,"%14s[%3d]", "alpha", j);
		for (i=1; i<=n_var; i++) fprintf(Fp_out, ",%11.3e", -Ad_inv[i][j]); 
		fprintf(Fp_out,"\n");
	}
	
	fprintf(Fp_out,"\n$ Sensitivities of Metabolites with respect to kinetic orders $\n");
	fprintf(Fp_out,"%-19s", "param/var");
	for (i=1; i<=n_var; i++) fprintf(Fp_out,",     y[%3d]", i); 
	fprintf(Fp_out,"\n");
	/* G ˴ؤ׻Ƚ񤭽Ф */
	sens_kinetic(Fp_out, Gd, Gi, Ad_inv, y, n_var, n_const, true);
	/* H ˴ؤ׻Ƚ񤭽Ф */
	sens_kinetic(Fp_out, Hd, Hi, Ad_inv, y, n_var, n_const, false);
	
	/* եåФ봶 */
	fprintf(Fp_out,"\n$ Logarythmic Gains of Fluxes $\n");
	fprintf(Fp_out,"%-19s", "param/flux");
	for (i=1; i<=n_var; i++) fprintf(Fp_out,",     v[%3d]", i); 
	fprintf(Fp_out,"\n");
	for (j=1; j<=n_const; j++) {
		fprintf(Fp_out,"%14s[%3d]", CONSTANT_PLAYER, j);
		for (i=1; i<=n_var; i++) fprintf(Fp_out, ",%11.3e", Gi[i][j]-GdAd1Ai[i][j]); 
		fprintf(Fp_out,"\n");
	}
	
	fprintf(Fp_out,"\n$ Sensitivities of Fluxes with respect to rate constants $\n");
	fprintf(Fp_out,"%-19s", "param/var");
	for (i=1; i<=n_var; i++) fprintf(Fp_out,",     v[%3d]", i); 
	fprintf(Fp_out,"\n");
	for (j=1; j<=n_var; j++) {
		fprintf(Fp_out,"%14s[%3d]", "alpha", j);
		for (i=1; i<=n_var; i++) {
			if (i == j) {
				val = 1.0;
			} else {
				val = 0.0;
			}
			for (k=1; k<=n_var; k++) val += Gd[i][k]*(-Ad_inv[k][j]);
			fprintf(Fp_out, ",%11.3e", val); 
		}
		fprintf(Fp_out,"\n");
	}
	
	fprintf(Fp_out,"\n$ Sensitivities of Fluxes with respect to kinetic orders $\n");
	fprintf(Fp_out,"%-19s", "param/var");
	for (i=1; i<=n_var; i++) fprintf(Fp_out,",     v[%3d]", i); 
	fprintf(Fp_out,"\n");
	/* G ˴ؤ׻Ƚ񤭽Ф */
	sens_kinetic_flux(Fp_out, Gd, Gi, Ad_inv, y, n_var, n_const, true);
	/* H ˴ؤ׻Ƚ񤭽Ф */
	sens_kinetic_flux(Fp_out, Hd, Hi, Ad_inv, y, n_var, n_const, false);

#ifdef DBG
	nr_dump_matrix("Gd", Gd, n_var, n_var);
	nr_dump_matrix("Hd", Hd, n_var, n_var);
	nr_dump_matrix("Gi", Gi, n_var, n_const);
	nr_dump_matrix("Hi", Hi, n_var, n_const);
	nr_dump_matrix("Ad_inv", Ad_inv, n_var, n_var);
#endif

	nr_free_vector(col, 1, n_var);
	nr_free_ivector(index, 1, n_var);
	nr_free_matrix(Ad, 1, n_var, 1, n_var);
	nr_free_matrix(Ad_inv, 1, n_var, 1, n_var);
	nr_free_matrix(Ai, 1, n_var, 1, n_const);
	nr_free_matrix(Ad1Ai, 1, n_var, 1, n_const);
	nr_free_matrix(GdAd1Ai, 1, n_var, 1, n_const);

	return 0;

err_alloc:
	log_sys("#  Cannot allocate memory.\n");

err:
	if (col) nr_free_vector(col, 1, n_var);
	if (beta_save) nr_free_vector(beta_save, 1, n_var);
	if (flux) nr_free_vector(flux, 1, n_var);
	if (index) nr_free_ivector(index, 1, n_var);
	if (Ad) nr_free_matrix(Ad, 1, n_var, 1, n_var);
	if (Ad_inv) nr_free_matrix(Ad_inv, 1, n_var, 1, n_var);
	if (Ai) nr_free_matrix(Ai, 1, n_var, 1, n_const);
	if (Ad1Ai) nr_free_matrix(Ad1Ai, 1, n_var, 1, n_const);
	if (GdAd1Ai) nr_free_matrix(GdAd1Ai, 1, n_var, 1, n_const);
	
	log_sys("## Error in std_analysis.\n");
	return -1;
}

/*---------------------- sens_kinetic -------------------------------------
 *	kinetic oder ФMetaboliteδ٤η׻Ƚ񤭽Ф
 */
static void
sens_kinetic(FILE *Fp_out, double **GHd, double **GHi, double **Ad_inv,
			   	double *y, int n_var, int n_const, int f_isG)
{
	int i, j, k, cnt;
	double *const_var = Param.val[1], my_sign;
	char label[MAX_LINE];

	if (f_isG) {
		sprintf(label, "%s", "g");
		my_sign = -1.0;
	} else {
		sprintf(label, "%s", "h");
		my_sign = 1.0;
	}

	for (j=1, cnt=0; j<=n_var; j++) {
		for (k=1; k<=n_var; k++) {
			if (GHd[j][k] == 0.0) continue;
			fprintf(Fp_out,"%14s[%3d]", label, ++cnt);
			for (i=1; i<=n_var; i++) {
				fprintf(Fp_out, ",%11.3e", my_sign*GHd[j][k]*Ad_inv[i][j]*log(y[k])); 
			}
			fprintf(Fp_out,"\n");
		}
		for (k=1; k<=n_const; k++) {
			if (GHi[j][k] == 0.0) continue;
			fprintf(Fp_out,"%14s[%3d]", label, ++cnt);
			for (i=1; i<=n_var; i++) {
				fprintf(Fp_out, ",%11.3e", my_sign*GHi[j][k]*Ad_inv[i][j]*log(const_var[k])); 
			}
			fprintf(Fp_out,"\n");
		}
	}
}

/*---------------------- sens_kinetic_flux -------------------------------------
 *	kinetic oder ФFluxδ٤η׻Ƚ񤭽Ф
 */
static void
sens_kinetic_flux(FILE *Fp_out, double **GHd, double **GHi, double **Ad_inv,
			   	double *y, int n_var, int n_const, int f_isG)
{
	int i, j, k, m, cnt;
	double *const_var = Param.val[1], val, my_sign;
	char label[MAX_LINE];

	if (f_isG) {
		sprintf(label, "%s", "g");
		my_sign = -1.0;
	} else {
		sprintf(label, "%s", "h");
		my_sign = 1.0;
	}

	for (j=1, cnt=0; j<=n_var; j++) {
		for (k=1; k<=n_var; k++) {
			if (GHd[j][k] == 0.0) continue;
			fprintf(Fp_out,"%14s[%3d]", label, ++cnt);
			for (i=1; i<=n_var; i++) {
				if (i == j) {
					val = GHd[j][k]*log(y[k]);
				} else {
					val = 0.0;
				}
				for (m=1; m<=n_var; m++) {
					val += GHd[i][m]*my_sign*GHd[j][k]*Ad_inv[m][j]*log(y[k]);
				}
				fprintf(Fp_out, ",%11.3e", val); 
			}
			fprintf(Fp_out,"\n");
		}
		for (k=1; k<=n_const; k++) {
			if (GHi[j][k] == 0.0) continue;
			fprintf(Fp_out,"%14s[%3d]", label, ++cnt);
			for (i=1; i<=n_var; i++) {
				if (i == j) {
					val = GHi[j][k]*log(const_var[k]);
				} else {
					val = 0.0;
				}
				for (m=1; m<=n_var; m++) {
					val += GHd[i][m]*my_sign*GHi[j][k]*Ad_inv[m][j]*log(const_var[k]);
#if 0
					if (f_isG == false) {
							printf("%d, %d, %d, %d, %e\n", i, j, k, m, val);
					}
#endif
				}
				fprintf(Fp_out, ",%11.3e", val); 
			}
			fprintf(Fp_out,"\n");
		}
	}
}

/*---------------------- allocate_base_matrixes -------------------------------------
 *	ܥޥȥåΰ
 */
static int
allocate_base_matrixes(double ***Gd, double ***Gi, double ***Hd, double ***Hi,
			    	int n_var, int n_const)
{
	*Gd = nr_alloc_matrix(1, n_var, 1, n_var);
	if (*Gd == NULL) goto err_alloc;
	*Gi = nr_alloc_matrix(1, n_var, 1, n_const);
	if (*Gi == NULL) goto err_alloc;
	*Hd = nr_alloc_matrix(1, n_var, 1, n_var);
	if (*Hd == NULL) goto err_alloc;
	*Hi = nr_alloc_matrix(1, n_var, 1, n_const);
	if (*Hi == NULL) goto err_alloc;

	return 0;

err_alloc:
	log_sys("## Error in allocate_base_matrixes.\n");
	return -1;
}

/*---------------------- free_base_matrixes -------------------------------------
 *	ܥޥȥåΰ賫
 */
static void
free_base_matrixes(double ***Gd, double ***Gi, double ***Hd, double ***Hi,
			    	int n_var, int n_const)
{
	if (*Gd) {
		nr_free_matrix(*Gd, 1, n_var, 1, n_var);
		*Gd = NULL;
	}
	if (*Gi) {
		nr_free_matrix(*Gi, 1, n_var, 1, n_var);
		*Gi = NULL;
	}
	if (*Hd) {
		nr_free_matrix(*Hd, 1, n_var, 1, n_var);
		*Hd = NULL;
	}
	if (*Hi) {
		nr_free_matrix(*Hi, 1, n_var, 1, n_var);
		*Hi = NULL;
	}
}
