#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <mpi.h>

#include "utl_mpi_ft.h"
#include "logUtil.h"

ftmpi_T	FtFileInfo;

/* prottypes of static functions */
static int get_filesize(char *file_name);


int SendFile(char *file_name, int dest)
{
	int i;
	int size;
	int rv = -1;
	FILE *fp = NULL;

#ifdef DBG
	DBGW(10, ("SendFile enter.\n"));
#endif /* DBG */
	
	if((size = get_filesize(file_name)) < 0){
		goto out;
	}

	FtFileInfo.fileSize = size;
	
	if(MPI_Send(&size, 1, MPI_INT, dest, 0, MPI_COMM_WORLD) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Send(In SendFile).\n");
		goto out;
	}
	
	/* send file */	
	if((FtFileInfo.buffer = (char *)malloc(size)) == NULL){
		log_sys(" ## Error: malloc(In SendFile).\n");
		goto out;
	}

	if((fp = fopen(file_name, "r")) == NULL){
		log_sys(" ## Error: Cannot open file %s\n", file_name);
		goto out;
	}

	for(i = 0; i < size; i++){
		FtFileInfo.buffer[i] = fgetc(fp);
	}

	/* Send buffer */
	if(MPI_Send(FtFileInfo.buffer, size, MPI_CHAR, 
							dest, 0, MPI_COMM_WORLD) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Send(In SendFile2).\n");
		goto out;
	}

	rv = 0;
out:
	if(fp != NULL) fclose(fp);
	if(FtFileInfo.buffer != NULL) free(FtFileInfo.buffer);

#ifdef DBG
	DBGW(10, ("SendFile exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

int RecvFile(char *file_name)
{
	int	i;
	int	rv = -1;
	FILE	*fp = NULL;
	MPI_Status stat;
	
#ifdef DBG
	DBGW(10, ("RecvFile enter\n"));
#endif /* DBG */

	if(MPI_Recv(&(FtFileInfo.fileSize), 1, MPI_INT, MPI_ANY_SOURCE, 
					MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Recv(In RecvFile).\n");
		goto out;
	}

	if((FtFileInfo.buffer = (char *)malloc(FtFileInfo.fileSize)) == NULL){
		log_sys(" ## Error: malloc(In RecvFile).\n");
		goto out;
	}
	
	/* recv file */
	if(MPI_Recv(FtFileInfo.buffer, FtFileInfo.fileSize, MPI_CHAR,
			MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &stat) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Recv(In RecvFile2).\n");
		goto out;
	}

	if((fp = fopen(file_name, "w")) == NULL){
		log_sys(" ## Error: Cannot open %s\n", file_name);
		goto out;
	}

	for(i = 0; i < FtFileInfo.fileSize; i++){
		if(putc(FtFileInfo.buffer[i], fp) == EOF){
			log_sys(" ## Error: putc(In RecvFile).\n");
			goto out;
		}
	}
	
	rv = 0;	
out:
	if(fp != NULL) fclose(fp);
	if(FtFileInfo.buffer != NULL) free(FtFileInfo.buffer);

#ifdef DBG
	DBGW(10, ("RecvFile exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

	

int BcastFileSend(char *file_name)
{
	int	i;
	int size;
	int	rv = -1;
	FILE *fp = NULL;
	
#ifdef DBG
	DBGW(10, ("BcastFileSend enter\n"));
#endif /* DBG */

	if((size = get_filesize(file_name)) < 0){
		goto out;
	}

	FtFileInfo.fileSize = size;

	if(MPI_Bcast(&size, 1, MPI_INT, 0, MPI_COMM_WORLD) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Bcast(In BcastFileSend).\n");
		goto out;
	}
	
	/* send file info all slave */	
	if((FtFileInfo.buffer = (char *)malloc(size)) == NULL){
		log_sys(" ## Error: malloc(In BcastFileSend).\n");
		goto out;
	}

	if((fp = fopen(file_name, "r")) == NULL){
		log_sys(" ## Error: Cannot open file %s\n", file_name);
		goto out;
	}

	for(i = 0; i < size; i++){
		FtFileInfo.buffer[i] = fgetc(fp);
	}

	/* Broadcast buffer */
	if(MPI_Bcast(FtFileInfo.buffer, size, MPI_CHAR, 
							0, MPI_COMM_WORLD) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Bcast(In BcastFileSend2).\n");
		goto out;
	}

	rv = 0;
out:
	if(fp != NULL) fclose(fp);
	if(FtFileInfo.buffer != NULL) free(FtFileInfo.buffer);

#ifdef DBG
	DBGW(10, ("BcastFileSend exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

int BcastFileRecv(char *file_name)
{
	int	i;
	int	rv = -1;
	FILE	*fp = NULL;

#ifdef DBG
	DBGW(10, ("BcastFileRecv enter\n"));
#endif /* DBG */

	if(MPI_Bcast(&(FtFileInfo.fileSize), 1, MPI_INT,
							0, MPI_COMM_WORLD) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Bcast(In BcastFileRecv).\n");
		goto out;
	}

	if((FtFileInfo.buffer = (char *)malloc(FtFileInfo.fileSize)) == NULL){
		log_sys(" ## Error: malloc(In BcastFileRecv).\n");
		goto out;
	}
	
	/* recv file */
	if(MPI_Bcast(FtFileInfo.buffer, FtFileInfo.fileSize, MPI_CHAR,
							0, MPI_COMM_WORLD) != MPI_SUCCESS){
		log_sys(" ## Error: MPI_Bcast(In BcastFileRecv2).\n");
		goto out;
	}

	if((fp = fopen(file_name, "w")) == NULL){
		log_sys(" ## Error: Cannot open %s\n", file_name);
		goto out;
	}

	for(i = 0; i < FtFileInfo.fileSize; i++){
		if(putc(FtFileInfo.buffer[i], fp) == EOF){
			log_sys(" ## Error: putc(In BcastFileRecv).\n");
			goto out;
		}
	}
	
	rv = 0;	
out:
	if(fp != NULL) fclose(fp);
	if(FtFileInfo.buffer != NULL) free(FtFileInfo.buffer);

#ifdef DBG
	DBGW(10, ("BcastFileRecv exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/***** Static Functions *****/

int get_filesize(char *file_name)
{
	int size;
	FILE *fp;
#ifdef DBG
	DBGW(10, ("get_filesize enter.\n"));
#endif /* DBG */
	
	if((fp = fopen(file_name, "r")) == NULL){
		log_sys(" ## Error: Cannot open %s\n", file_name);
		goto out;
	}

	if(fseek(fp, 0, SEEK_END) != 0){
		log_sys(" ## Error: fseek(In get_filesize).\n");
		goto out;
	}
	
	if((size = ftell(fp)) < 0){
		log_sys(" ## Error: ftell(In get_filesize).\n");
		goto out;
	}

out:
	if(fp != NULL) fclose(fp);

#ifdef DBG
	DBGW(10, ("get_filesize exit(%d)\n", size));
#endif /* DBG */

	return size;
}

#ifdef TEST
int main(int argc, char **argv)
{
	int my_rank;
	int	rv = EXIT_ERROR;
	char file_name[FILENAME_MAX];

	MPI_Init(&argc, &argv);

	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

	if(my_rank == 0){
		if(SendFileAllSlave(argv[1]) == EXIT_ERROR){
			goto out;
		}
	} else {
	 	sprintf(file_name, "%s_%d", argv[1], my_rank);
		if(RecvFile(file_name) == EXIT_ERROR){
			goto out;
		}
		if(unlink(file_name) < 0){
			perror("unlink");
			goto out;
		}
	}

	rv = EXIT_NORMAL;
out:
	MPI_Finalize();
	return rv;
}



#endif /* TEST */
